import logging
import os
import os.path
from urllib.parse import urlencode

from filelock import FileLock, Timeout
from jinja2 import Template
from pkg_resources import parse_version

from packy_agent.configuration.settings import settings
from packy_agent.exceptions import AuthenticationError, ValidationError
from packy_agent.domain_logic.managers.control import control_manager
from packy_agent.domain_logic.constants import WORKER_COMPONENT
from packy_agent.utils.pkg_resources import get_package_file_content
from packy_agent.utils.shell import run_command_async
from packy_agent.utils.output import write_to_console_or_file
from packy_agent.clients.packy_server import get_packy_server_client


PACKY_GET_INSTALL_SCRIPT_PATH = '/tmp/packy-run-install-script.sh'


logger = logging.getLogger(__name__)


def render_template_from_package_file_content(template_module_name, template_filename,
                                              context=None):
    template = Template(
        get_package_file_content(template_module_name, template_filename).decode('utf-8'))
    return template.render(**(context or {}))


def validate_version(version):
    version_max = get_packy_server_client().get_version_max()
    if version_max is None:
        return version

    if version is None:
        return version_max

    if parse_version(version) > parse_version(version_max):
        raise ValueError(f'Could not install/upgrade to version {version} '
                         f'(it is higher than {version_max})')

    return version


def generate_packy_run_install_script(script_filename, version):
    context = {'with_shebang': True}
    if version:
        context['download_args'] = '?{}'.format(urlencode({'version': version}))

    write_to_console_or_file(
        script_filename,
        render_template_from_package_file_content(
            'packy_agent', 'scripts/misc/packy-run-install-script.sh.j2', context))


def run_packy_run_install_script(script_filename, upgrade_in_progress_lock_filename, delay_seconds):
    extra_envvars = {
        'PACKY_SERVER_BASE_URL': settings.get_server_base_url(),
        'PACKY_AGENT_ACCESS_TOKEN': settings.get_access_token(),
    }

    old_path = os.getenv('OLD_PATH')
    if old_path:
        extra_envvars['PATH'] = old_path

    export_envvars = ' '.join(f"export {k}='{v}';" for k, v in extra_envvars.items())
    command = (f'{export_envvars} /bin/bash {script_filename}; '
               f'rm -f {upgrade_in_progress_lock_filename}')

    run_command_async(command, delay_seconds=delay_seconds)


def refresh_access_token():
    old_access_token = settings.get_access_token()

    tokens = get_packy_server_client().refresh_access_token(
        settings.get_client_id(), settings.get_refresh_token())
    settings.set_access_token(tokens['access_token'])
    settings.set_refresh_token(tokens['refresh_token'])

    invalid_access_tokens = settings.get_invalid_access_tokens()
    if invalid_access_tokens:
        invalid_access_tokens.discard(old_access_token)


class InstallAndUpgradeManager:

    def activate(self, email=None, password=None, access_token=None, agent_id=None,
                 ensure_restarted=True):
        if (email is None or password is None) and access_token is None:
            raise ValueError('Either email and password or access_token must be provided')

        if settings.is_activated():
            raise ValidationError('Agent has already been activated')

        if agent_id is None:
            response = get_packy_server_client().create_agent(email, password, raise_for_status=False)
        elif access_token is None:
            response = get_packy_server_client().activate_agent_basic_auth(
                email, password, agent_id, raise_for_status=False)
        else:
            response = get_packy_server_client().activate_agent_token_auth(
                access_token, agent_id, raise_for_status=False)

        status_code = response.status_code
        if status_code == 401:
            raise AuthenticationError('Invalid credentials for agent activation')
        elif status_code == 400:
            payload = response.json()
            errors = payload.get('errors') or {}
            non_field_errors = errors.pop('non_field_errors', None) or []
            messages = ([payload.get('message')] +
                        [nfe.get('message') for nfe in non_field_errors] +
                        ['{}: {}.'.format(k, v.get('message')) for k, v in errors.items()])

            raise ValidationError(' '.join(filter(None, messages)))

        response.raise_for_status()

        response_json = response.json()
        settings.activate(
            client_id=response_json['client_id'],
            access_token=response_json['access_token'],
            refresh_token=response_json['refresh_token'],
            agent_key=response_json['key'],
            agent_name=response_json['name'],
        )

        if ensure_restarted:
            if control_manager.is_running(WORKER_COMPONENT):
                control_manager.restart(WORKER_COMPONENT)
            else:
                control_manager.start(WORKER_COMPONENT)

    def install_and_restart(self, version=None, delay_seconds=None):
        logger.info('Upgrading...')
        if not settings.is_upgrade_enabled():
            logger.info('Upgrade was disabled (for developer protection)')
            return

        upgrade_in_progress_lock_filename = settings.get_upgrade_in_progress_lock_filename()
        file_lock = FileLock(upgrade_in_progress_lock_filename)
        try:
            with file_lock.acquire(timeout=1):
                version = validate_version(version)

                script_filename = settings.get_packy_run_install_script_filename()
                generate_packy_run_install_script(script_filename, version)
                run_packy_run_install_script(script_filename, upgrade_in_progress_lock_filename,
                                             delay_seconds)
        except Timeout:
            logger.warning('Concurrent upgrade detected')

    def upgrade_and_restart(self, delay_seconds=None):
        self.install_and_restart(delay_seconds=delay_seconds)


# As long as manager is stateless object we do not need to care about greenlet-multitasking safety
install_and_upgrade_manager = InstallAndUpgradeManager()



from abc import ABC
import logging

from gevent.event import Event

from packy_agent.worker.loops.base.loop import Loop


logger = logging.getLogger(__name__)


class StoppableLoop(Loop, ABC):

    loop_type = 'stoppable'

    def __init__(self, is_safe_iteration=True, is_logged_iteration=True):
        super().__init__(is_safe_iteration=is_safe_iteration,
                         is_logged_iteration=is_logged_iteration)
        self.is_running = False
        self.is_stopping = False
        self.wake_up_event = Event()

    def loop(self):
        self.is_running = True
        self.is_stopping = False
        logger.debug(f'STARTED {self.description}')

        while self.is_running:
            self.iteration_wrapper()
            if self.is_stopping:
                break

        self.is_running = False
        self.is_stopping = False
        logger.debug(f'STOPPED {self.description}')

    def sleep(self, timeout):
        self.wake_up_event.clear()
        self.wake_up_event.wait(timeout)

    def stop(self):
        logger.debug(f'STOP requested for {self.description}')
        self.is_stopping = True
        self.wake_up_event.set()



import logging
import sys
from getpass import getpass

from packy_agent.clients.packy_server import get_inactive_agents
from packy_agent.domain_logic.managers.install_and_upgrade import install_and_upgrade_manager
from packy_agent.exceptions import AuthenticationError, ValidationError
from packy_agent.configuration.settings import settings
from packy_agent.utils.cli import get_base_argument_parser
from packy_agent.utils.logging import configure_logging, configure_logging_basic


def entry():
    configure_logging_basic(logging.WARNING)
    parser = get_base_argument_parser(__loader__.name)

    parser.add_argument('--email')
    parser.add_argument('--password')

    args = parser.parse_args()
    settings.set_commandline_arguments(vars(args))
    configure_logging(logging.WARNING)

    if settings.is_activated():
        print('Packy Agent is already activated')
        return 1

    print('Packy Agent activation')

    email = args.email or input('E-mail: ')
    password = args.password or getpass('Password: ')

    try:
        agents = get_inactive_agents(email, password)
        activating_agent_id = None
        if agents:
            print('Activate as:')
            print('[enter] New agent')
            for agent_id, agent_name in agents.items():
                print(f'[id={agent_id}] {agent_name}')

            while True:
                agent_id = input('Agent id: ') or None
                if agent_id:
                    try:
                        activating_agent_id = int(agent_id)
                    except (ValueError, TypeError):
                        print('Agent ID must be an integer')
                    else:
                        break
                else:
                    break

        install_and_upgrade_manager.activate(email, password, agent_id=activating_agent_id)
    except AuthenticationError:
        print('Not authenticated (invalid credentials)')
        return 1
    except ValidationError as ex:
        print(f'Validation error: {ex}')
        return 1

    print()
    print('-------------------------------------')
    print('Packy Agent is activated successfully')
    print()


if __name__ == '__main__':
    sys.exit(entry())



import pytest

from packy_agent.clients.sentry import init_sentry_client


@pytest.fixture(scope='session', autouse=True)
def sentry_server_mock(test_config):
    events = []

    def capture(event):
        events.append(event)

    init_sentry_client(packy_agent_sentry_dsn='https://dummy:dummy@dummy.dummy/1',
                       transport=capture)

    return events



import pytest

from packy_agent.console.app import get_app


@pytest.fixture
def app():
    return get_app()


@pytest.fixture
def console_flask_client(client, config):
    testing = config['TESTING']
    config['TESTING'] = True
    wtf_csrf_enabled = config.get('WTF_CSRF_ENABLED', True)
    config['WTF_CSRF_ENABLED'] = False
    yield client
    config['TESTING'] = testing
    config['WTF_CSRF_ENABLED'] = wtf_csrf_enabled



from schematics import types

from packy_agent.domain_logic.models.schematics.base import BaseMeasurement


class HTTPModuleMeasurement(BaseMeasurement):
    target = types.StringType(required=True)

    is_success = types.BooleanType(default=True)
    http_status_code = types.IntType()

    namelookup_ms = types.FloatType()
    total_ms = types.FloatType()

    certificate_expiration_dt = types.DateTimeType()



import gevent

from packy_agent.worker.loops.base.periodic import PeriodicLoop


def test_periodic_loop_survives_exception_in_callable():
    counter = 0
    with_exception = False

    def call_me():
        nonlocal counter
        if with_exception:
            raise Exception('Test exception')
        counter += 1

    loop = PeriodicLoop(0.5, call_me)
    loop.start()
    gevent.sleep(1)
    with_exception = True
    counter_value_before_exception = counter
    assert counter_value_before_exception
    gevent.sleep(1)
    with_exception = False
    gevent.sleep(1)
    assert counter > counter_value_before_exception
    loop.stop()
    gevent.sleep(1)



import logging
import socket
import os

import gevent
from ryu.lib.packet.icmp import icmp, ICMP_ECHO_REPLY, ICMP_TIME_EXCEEDED, ICMP_DEST_UNREACH
from ryu.lib.packet.ipv4 import ipv4
from ryu.lib.packet.packet import Packet

from packy_agent.domain_logic.modules.common.socket import (
    get_seq_no, send_icmp_echo_request as send_icmp_probe, decode_timestamp, ICMP_PROTOCOL,
    TIME_SIZE_BYTES)
from packy_agent.domain_logic.modules.traceroute.base import Traceroute
from packy_agent.domain_logic.modules.traceroute.models import SentProbe
from packy_agent.domain_logic.modules.common.constants import (IP_PACKET_HEADER_SIZE_BYTES,
                                                               ICMP_PACKET_HEADER_SIZE_BYTES)
from packy_agent.utils.misc import safe_hexlify

PACKET_HEADERS_SIZE_BYTES = IP_PACKET_HEADER_SIZE_BYTES + ICMP_PACKET_HEADER_SIZE_BYTES

logger = logging.getLogger(__name__)


def get_icmp_id():
    return (os.getpid() + id(gevent.getcurrent())) & 0xFFFF


class TracerouteICMP(Traceroute):

    range_min = PACKET_HEADERS_SIZE_BYTES

    def send_probe(self, probe_number, destination_ip_address, ttl, packet_size_bytes=60, **kwargs):
        sending_socket = self.get_sending_socket()
        sending_socket.setsockopt(socket.SOL_IP, socket.IP_TTL, ttl)

        seq_no = get_seq_no(ICMP_PROTOCOL)
        sent_time, sent_payload = send_icmp_probe(
            sending_socket, destination_ip_address, get_icmp_id(), seq_no,
            payload_size_bytes=packet_size_bytes - PACKET_HEADERS_SIZE_BYTES, **kwargs)
        return SentProbe(ttl, probe_number, sent_time, sent_payload, seq_no)

    @staticmethod
    def is_probe_reply(icmp_packet):
        return icmp_packet.type in (ICMP_TIME_EXCEEDED, ICMP_ECHO_REPLY, ICMP_DEST_UNREACH)

    @staticmethod
    def retrieve_probe_packet(icmp_packet):
        if icmp_packet.type in (ICMP_TIME_EXCEEDED, ICMP_DEST_UNREACH):
            inner_packet = Packet(icmp_packet.data.data, parse_cls=ipv4)
            protocol_1 = inner_packet.protocols[1]
            if isinstance(protocol_1, icmp):
                effective_icmp_packet = protocol_1
            else:
                logger.debug('ICMP packet does not contain inner ICMP packet')
                return
        elif icmp_packet.type == ICMP_ECHO_REPLY:
            effective_icmp_packet = icmp_packet
        else:
            raise ValueError(f'Unsupported ICMP packet type: {icmp_packet.type}')

        logger.debug(
            'Processing ICMP packet id:%s, seq:%s, payload:%s', effective_icmp_packet.data.id,
            effective_icmp_packet.data.seq, safe_hexlify(effective_icmp_packet.data.data))

        return effective_icmp_packet

    def process_method_specific_probe_reply(self, ip_packet, icmp_packet, probe_packet):
        id_ = probe_packet.data.id
        if id_ != get_icmp_id():
            logger.debug('Discarded ICMP packet (reason: unexpected id %s) '
                         'from: %s, IP checksum: %s, type: %s, code: %s',
                         probe_packet.data.id,
                         ip_packet.src, ip_packet.csum, icmp_packet.type, icmp_packet.code)
            return False

        seq = probe_packet.data.seq
        if seq not in self.sent_probes:
            logger.debug('Discarded ICMP packet (reason: unexpected seq %s) '
                         'from: %s, IP checksum: %s, type: %s, code: %s', seq,
                         ip_packet.src, ip_packet.csum, icmp_packet.type, icmp_packet.code)
            return False

        return True

    @staticmethod
    def retrieve_payload(icmp_packet, probe_packet):
        return probe_packet.data.data

    def get_sent_probe(self, probe_packet):
        return self.sent_probes.get(probe_packet.data.seq)

    @staticmethod
    def is_last_hop(icmp_packet):
        if icmp_packet.type == ICMP_TIME_EXCEEDED:
            return False
        elif icmp_packet.type in (ICMP_ECHO_REPLY, ICMP_DEST_UNREACH):
            return True
        else:
            raise ValueError(
                f'Unsupported ICMP packet type: {icmp_packet.type} (should never get here)')

    @staticmethod
    def retrieve_sent_time_from_payload(payload):
        if payload and len(payload) >= TIME_SIZE_BYTES:
            return decode_timestamp(payload[:TIME_SIZE_BYTES])

    @staticmethod
    def get_hop_ip_address(ip_packet, icmp_packet):
        return None if icmp_packet.type == ICMP_DEST_UNREACH else ip_packet.src

    def open_sending_socket(self):
        logger.debug('Opening sending socket...')
        try:
            return socket.socket(socket.AF_INET, socket.SOCK_RAW, self.icmp_protocol)
        except socket.error as ex:
            if ex.errno == 1:  # Operation not permitted
                # TODO(dmu) HIGH: Fix to receive ICMP packet without being root
                raise socket.error(
                    ex.strerror + '  - process should be running as root.')

            raise

    def get_sending_socket(self):
        if not self.sending_socket:
            self.sending_socket = self.receiving_socket or self.open_sending_socket()

        return self.sending_socket

    def get_receiving_socket(self):
        if not self.receiving_socket:
            self.receiving_socket = self.sending_socket or self.open_receiving_socket()

        return self.receiving_socket



import logging

logger = logging.getLogger(__name__)


def write_to_console_or_file(target, content):
    if target == '-':
        print(content)
    else:
        logger.info('Writing content to %s:\n%s', target, content)
        # TODO(dmu) MEDIUM: Consider using atomic writes
        with open(target, 'w') as f:
            f.write(content)
        logger.info('Written content to %s', target)



import logging

from sentry_sdk import capture_message
from requests.exceptions import ConnectionError

from packy_agent.constants import (
    PING_MODULE, TRACE_MODULE, SPEEDTEST_MODULE, HTTP_MODULE, SUBMITTER_LOOP, ErrorSide)
from packy_agent.worker.loops.base.periodic import PeriodicLoop
from packy_agent.clients.local_storage import get_local_storage, TYPE_MAP
from packy_agent.configuration.settings import settings
from packy_agent.clients.packy_server import get_packy_server_client
from packy_agent.exceptions import CoolDown


MODULE_PUBLIC_IDENTIFIER_MAP = {
    1: PING_MODULE,
    2: TRACE_MODULE,
    3: SPEEDTEST_MODULE,
    4: HTTP_MODULE,
}
assert len(MODULE_PUBLIC_IDENTIFIER_MAP) == len(TYPE_MAP)

logger = logging.getLogger(__name__)


def is_invalid_chronological_order(response):
    if not response.status_code == 400:
        return False

    response_json = response.json()
    if not isinstance(response_json, dict):
        return False

    errors = response_json.get('errors')
    if not isinstance(errors, dict):
        return False

    non_field_errors = errors.get('non_field_errors')
    if not non_field_errors or not isinstance(non_field_errors, list):
        return False

    for item in non_field_errors:
        if item.get('code') == 'invalid-chronological-order':
            return True

    return False


def submit_measurement(measurement_row):
    module_public_identifier = MODULE_PUBLIC_IDENTIFIER_MAP.get(measurement_row.measurement_type)
    if not module_public_identifier:
        logger.warning(f'Unknown measurement type: {measurement_row.measurement_type}')
        return

    local_storage = get_local_storage()
    value = measurement_row.value
    capture_message(f'Got measurement value from database: {value!r}')
    try:
        response = get_packy_server_client().submit_measurement(
            module_public_identifier, value, raise_for_status=False)
    except ConnectionError as ex:
        logger.warning(f'Could not connect to server to submit measurement of '
                       f'{module_public_identifier} ({ex!r}): {value}')
        local_storage.save_measurement_error(
            measurement_row.id, ErrorSide.SERVER.value, f'Could not connect to server: {ex!r}')
        raise CoolDown()
    except Exception as ex:
        logger.error(f'Error while submitting measurement of {module_public_identifier}: {value}')
        local_storage.save_measurement_error(
            measurement_row.id, ErrorSide.CLIENT.value, repr(ex) or 'Unknown')
        raise CoolDown()

    status_code = response.status_code
    if 200 <= status_code < 300:
        local_storage.mark_measurement_as_submitted(measurement_row.id)
        return
    else:
        if status_code >= 500:
            error_side = ErrorSide.SERVER.value
            log_method = logger.warning  # because it is server error
            cool_down = True
        else:
            error_side = ErrorSide.CLIENT.value
            if is_invalid_chronological_order(response):
                # Marking as submitted is actually a kind of hack
                local_storage.mark_measurement_as_submitted(measurement_row.id)
                log_method = logger.warning
                cool_down = False
            else:
                log_method = logger.error  # because it is our error
                cool_down = True

        log_method(f'Server responded with HTTP{status_code}: {response.content}')
        local_storage.save_measurement_error(
            measurement_row.id, error_side, f'HTTP{status_code}: {response.content}')

        if cool_down:
            raise CoolDown()


class TaskResultsSubmitter(PeriodicLoop):

    formal_name = SUBMITTER_LOOP

    def __init__(self):
        super().__init__(settings.get_worker_results_submission_period_seconds())
        self.submitted_counter = 0

    def call(self):
        results_submission_pause_seconds = settings.get_worker_results_submission_pause_seconds()
        for measurement_type, module_name in MODULE_PUBLIC_IDENTIFIER_MAP.items():
            logger.debug('Submitting measurements for %s module...', module_name)
            for measurement_row in get_local_storage().get_remaining_measurements(measurement_type):
                try:
                    submit_measurement(measurement_row)
                    self.submitted_counter += 1
                except CoolDown:
                    break
                finally:
                    self.sleep(results_submission_pause_seconds)
                    if self.is_stopping or not self.is_running:
                        return



import time
from unittest.mock import MagicMock

import pytest
import httpretty

from packy_agent.configuration.settings import settings
from packy_agent.watchdog.service import WatchdogService
from packy_agent.tests.fixtures.configuration import override_settings


@pytest.mark.usefixtures('activated', 'watchdog_started_long_ago')
def test_online_status_for_restart():
    with override_settings(
            {'worker': {'offline_to_restart_seconds': 10, 'restart_wait_seconds': 5}}):
        with httpretty.enabled():
            url = settings.get_server_base_url() + 'api/v2/agent/'
            httpretty.register_uri(httpretty.GET, url, status=200, body='{"is_online": false}')
            httpretty.register_uri(httpretty.GET, settings.get_console_base_url(),
                                   status=200)  # emulate console running

            service = WatchdogService()
            service.last_known_online_ts = time.time() - 11

            service.reboot = MagicMock()
            service.restart_worker = MagicMock()
            service.resolve_worker_supervisor_program_status = MagicMock()
            service.resolve_worker_process_status = MagicMock()
            service.resolve_worker_response = MagicMock()

            service.loop_iteration()

            service.restart_worker.assert_called()
            service.reboot.assert_not_called()


@pytest.mark.usefixtures('activated', 'watchdog_started_long_ago')
def test_online_status_for_reboot():
    now_ts = time.time()
    with override_settings(
            {'worker': {'offline_to_reboot_seconds': 10, 'reboot_wait_seconds': 5},
             'rebooted_at_ts': now_ts - 12}):
        with httpretty.enabled():
            url = settings.get_server_base_url() + 'api/v2/agent/'
            httpretty.register_uri(httpretty.GET, url, status=200, body='{"is_online": false}')
            httpretty.register_uri(httpretty.GET, settings.get_console_base_url(),
                                   status=200)  # emulate console running

            service = WatchdogService()
            service.last_known_online_ts = time.time() - 11

            service.reboot = MagicMock()
            service.restart_worker = MagicMock()
            service.resolve_worker_supervisor_program_status = MagicMock()
            service.resolve_worker_process_status = MagicMock()
            service.resolve_worker_response = MagicMock()

            service.loop_iteration()

            service.reboot.assert_called()
            service.restart_worker.assert_not_called()


@pytest.mark.usefixtures('activated', 'watchdog_started_long_ago')
def test_online_status_for_ok():
    with httpretty.enabled():
        url = settings.get_server_base_url() + 'api/v2/agent/'
        httpretty.register_uri(httpretty.GET, url, status=200, body='{"is_online": true}')
        httpretty.register_uri(httpretty.GET, settings.get_console_base_url(),
                               status=200)  # emulate console running

        service = WatchdogService()
        service.last_known_online_ts = time.time() - 11

        service.reboot = MagicMock()
        service.restart_worker = MagicMock()
        service.resolve_worker_supervisor_program_status = MagicMock()
        service.resolve_worker_process_status = MagicMock()
        service.resolve_worker_response = MagicMock()

        service.loop_iteration()

        service.reboot.assert_not_called()
        service.restart_worker.assert_not_called()



import logging

from packy_agent.exceptions import ImproperlyConfiguredError
from packy_agent.constants import SPEEDTEST_MODULE, SPEEDTEST_MODULE_LOOP
from packy_agent.worker.loops.modules.base import ScheduledProducerTaskLoop, get_target
from packy_agent.domain_logic.models.schematics.speedtest import SpeedtestModuleMeasurement
from packy_agent.domain_logic.modules.speedtest.base import run_speedtest

MBIT = 1000000.0


logger = logging.getLogger(__name__)


def speedtest_impl(*args, **kwargs):
    target = get_target(kwargs, SPEEDTEST_MODULE)
    if not target:
        return

    try:
        target_id = int(target)
    except (ValueError, TypeError):
        raise ImproperlyConfiguredError(f'Speedtest target must be a number (not {target!r})')

    # we place it here because it also saves time momen
    measurement = SpeedtestModuleMeasurement({'target': target})

    results = run_speedtest(target_id)
    measurement.upload_speed = round(results.upload / MBIT, 2)
    measurement.download_speed = round(results.download / MBIT, 2)
    measurement.ping = round(results.ping, 3)

    return measurement


class SpeedtestTaskLoop(ScheduledProducerTaskLoop):

    formal_name = SPEEDTEST_MODULE_LOOP

    def __init__(self, schedule, args=(), kwargs=None, outbound_queue=None):
        super().__init__(schedule=schedule, callable_=speedtest_impl, args=args, kwargs=kwargs,
                         outbound_queue=outbound_queue)



import sys
import logging

from packy_agent.configuration.settings import settings
from packy_agent.utils.cli import get_base_argument_parser
from packy_agent.utils.logging import configure_logging, configure_logging_basic

from packy_agent.domain_logic.modules.http.base import get_http_measurement


def entry():
    configure_logging_basic(logging.WARNING)
    parser = get_base_argument_parser(__loader__.name)

    parser.add_argument('url')
    parser.add_argument('--follow-redirects', action='store_true')

    args = parser.parse_args()
    settings.set_commandline_arguments(vars(args))
    configure_logging(logging.WARNING)

    url = args.url
    print(f'Getting information for {url}')
    print(get_http_measurement(url, follow_redirects=args.follow_redirects).to_json())


if __name__ == '__main__':
    sys.exit(entry())



import psutil


def get_network_usage():
    snetio = psutil.net_io_counters(pernic=True, nowrap=True)
    snetio.pop('lo', None)

    bytes_sent = bytes_received = 0
    for value in snetio.values():
        bytes_sent += value.bytes_sent
        bytes_received += value.bytes_recv

    return bytes_sent, bytes_received



import datetime
import time

import dateutil
import pytz
from croniter import croniter

EPOCH_START_DATETIME = datetime.datetime(1970, 1, 1, tzinfo=pytz.utc)


def get_iso_format_utcnow():
    return datetime.datetime.utcnow().isoformat() + 'Z'


def to_epoch(dt):
    return (dt - EPOCH_START_DATETIME).total_seconds()


def iso_to_datetime(dt_string):
    return dateutil.parser.parse(dt_string)


def get_croniter(schedule):
    schedule = schedule.split()
    schedule.append(schedule.pop(0))
    return croniter(' '.join(schedule), time.time())



from unittest.mock import patch, call

from packy_agent.domain_logic.managers.control import control_manager


def test_start():
    with patch('packy_agent.domain_logic.managers.control.run_supervisor_command') as mock:
        mock.side_effect = ('nginx RUNNING pid 11, uptime 1 minute', None)
        control_manager.start('worker')
        mock.assert_has_calls((call('start', ('worker',)),))


def test_start_nginx():
    with patch('packy_agent.domain_logic.managers.control.run_supervisor_command') as sup_mock:
        with patch('packy_agent.domain_logic.managers.control.nginx_service') as nginx_mock:
            with patch('packy_agent.domain_logic.managers.control.'
                       'is_inside_docker_container') as is_inside_docker_container_mock:
                is_inside_docker_container_mock.return_value = False
                nginx_mock.is_running.return_value = False
                control_manager.start('nginx')
                nginx_mock.start.assert_called()
                sup_mock.assert_not_called()


def test_start_nginx_in_docker():
    with patch('packy_agent.domain_logic.managers.control.run_supervisor_command') as sup_mock:
        with patch('packy_agent.domain_logic.managers.control.nginx_service') as nginx_mock:
            with patch('packy_agent.domain_logic.managers.control.'
                       'is_inside_docker_container') as is_inside_docker_container_mock:
                sup_mock.side_effect = ('nginx STOPPED', None)
                is_inside_docker_container_mock.return_value = True
                nginx_mock.is_running.return_value = False
                control_manager.start('nginx')
                sup_mock.assert_has_calls((call('start', ('nginx',)),), any_order=True)



from enum import Enum

PING_MODULE = 'ping'
TRACE_MODULE = 'trace'
SPEEDTEST_MODULE = 'speedtest'
HTTP_MODULE = 'http'

MODULES = (PING_MODULE, TRACE_MODULE, SPEEDTEST_MODULE, HTTP_MODULE)

MODULE_LOOP_SUFFIX = '_module'
PING_MODULE_LOOP = PING_MODULE + MODULE_LOOP_SUFFIX
TRACE_MODULE_LOOP = TRACE_MODULE + MODULE_LOOP_SUFFIX
SPEEDTEST_MODULE_LOOP = SPEEDTEST_MODULE + MODULE_LOOP_SUFFIX
HTTP_MODULE_LOOP = HTTP_MODULE + MODULE_LOOP_SUFFIX
CONSUMER_LOOP = 'consumer'
SUBMITTER_LOOP = 'submitter'
PURGER_LOOP = 'purger'
COMMUNICATION_LOOP = 'communication'
HEARTBEAT_LOOP = 'heartbeat'
GUARD_LOOP = 'guard'
STATUS_LOOP = 'status'

LOOPS = (PING_MODULE_LOOP, TRACE_MODULE_LOOP, SPEEDTEST_MODULE_LOOP, HTTP_MODULE_LOOP,
         CONSUMER_LOOP, SUBMITTER_LOOP, PURGER_LOOP, COMMUNICATION_LOOP, HEARTBEAT_LOOP,
         GUARD_LOOP, STATUS_LOOP)

RELOAD_EVENT = 'reload'
RESTART_EVENT = 'restart'
REBOOT_EVENT = 'reboot'
UPGRADE_EVENT = 'upgrade'
HEARTBEAT_EVENT = 'heartbeat'


class ErrorSide(Enum):
    SERVER = 1
    CLIENT = 2



import json
from collections import MutableMapping

from sqlitedict import SqliteDict

from packy_agent.configuration.sources.base import (
    ConfigurationSourceBase, READ_ONLY_EMPTY_DICT)
from packy_agent.utils.misc import get_lazy_value
from packy_agent.utils.gevent import database_locks


def key_value_encode(obj):
    return json.dumps(obj)


def key_value_decode(obj):
    return json.loads(obj)


class KeyValueStorageBackedSource(ConfigurationSourceBase, MutableMapping):

    is_writeable = True

    def __init__(self, tablename=None, filename=None, name=None):
        self._tablename = tablename
        self._filename = filename

        self._mapping = None

        super().__init__(name=name)

    def __getitem__(self, item):
        with database_locks[self.filename]:
            with self.get_mapping() as mapping:
                return mapping[item]

    def __iter__(self):
        with database_locks[self.filename]:
            with self.get_mapping() as mapping:
                yield from mapping

    def __len__(self):
        with database_locks[self.filename]:
            with self.get_mapping() as mapping:
                return len(mapping)

    def __contains__(self, item):
        with database_locks[self.filename]:
            with self.get_mapping() as mapping:
                return item in mapping

    def __eq__(self, other):
        with database_locks[self.filename]:
            with self.get_mapping() as mapping:
                return mapping == other

    def __ne__(self, other):
        with database_locks[self.filename]:
            with self.get_mapping() as mapping:
                return mapping != other

    def __setitem__(self, key, value):
        with database_locks[self.filename]:
            with self.get_mapping() as mapping:
                mapping[key] = value

    def __delitem__(self, key):
        with database_locks[self.filename]:
            with database_locks[self.filename]:
                with self.get_mapping() as mapping:
                    del mapping[key]

    def keys(self):
        with database_locks[self.filename]:
            with self.get_mapping() as mapping:
                yield from mapping.keys()

    def items(self):
        with database_locks[self.filename]:
            with self.get_mapping() as mapping:
                yield from mapping.items()

    def values(self):
        with database_locks[self.filename]:
            with self.get_mapping() as mapping:
                yield from mapping.values()

    def get(self, key, default=None):
        with database_locks[self.filename]:
            with self.get_mapping() as mapping:
                return mapping.get(key, default)

    def pop(self, key):
        with database_locks[self.filename]:
            with self.get_mapping() as mapping:
                return mapping.pop(key)

    def popitem(self):
        with database_locks[self.filename]:
            with self.get_mapping() as mapping:
                return mapping.popitem()

    def clear(self):
        with database_locks[self.filename]:
            with self.get_mapping() as mapping:
                mapping.clear()

    def update(self, *args, **kwargs):
        with database_locks[self.filename]:
            with self.get_mapping() as mapping:
                mapping.update(*args, **kwargs)

    def setdefault(self, *args, **kwargs):
        with database_locks[self.filename]:
            with self.get_mapping() as mapping:
                return mapping.setdefault(*args, **kwargs)

    def replace(self, new_mapping):
        with database_locks[self.filename]:
            with self.get_mapping() as mapping:
                mapping.clear()
                mapping.update(new_mapping)

    @property
    def is_ready(self):
        return (isinstance(self._tablename, str) and isinstance(self._filename, str) and
                self._mapping is not None)

    def ensure_ready(self):
        if self.is_ready or self.is_preparing:
            return

        self.get_mapping()

    @property
    def tablename(self):
        self._tablename = get_lazy_value(self._tablename)
        return self._tablename

    @property
    def filename(self):
        self._filename = get_lazy_value(self._filename)
        return self._filename

    def get_mapping(self):
        if self._mapping is None and not self._is_preparing:
            self._is_preparing = True
            try:
                filename = self.filename
                tablename = self.tablename
                if filename and tablename:
                    self._mapping = SqliteDict(
                        filename, tablename=tablename, autocommit=True, encode=key_value_encode,
                        decode=key_value_decode)
            finally:
                self._is_preparing = False

        return READ_ONLY_EMPTY_DICT if self._mapping is None else self._mapping



import logging

from packy_agent.domain_logic.managers.control import control_manager
from packy_agent.domain_logic.managers.network import network_manager
from packy_agent.configuration.settings import settings
from packy_agent.clients.packy_server import get_packy_server_client


logger = logging.getLogger(__name__)


class ResetManager:

    def reset_network_configuration(self, reboot=True, reboot_delay_seconds=5):
        if network_manager.is_network_configuration_supported():
            network_manager.restore_configuration(
                reboot=reboot, reboot_delay_seconds=reboot_delay_seconds)
        else:
            logger.debug('Network configuration is not supported')

    def deactivate(self, reboot=False, delay_seconds=5):
        get_packy_server_client().deactivate_agent()
        settings.deactivate()

        if reboot:
            control_manager.reboot(delay_seconds=delay_seconds)
        else:
            control_manager.restart_all(delay_seconds=delay_seconds)

    def full_reset(self, delay_seconds=5):
        self.deactivate(delay_seconds=delay_seconds)
        self.reset_network_configuration(reboot=True, reboot_delay_seconds=delay_seconds)


# As long as manager is stateless object we do not need to care about greenlet-multitasking safety
reset_manager = ResetManager()



import logging

import packy_agent
from packy_agent.configuration.settings import settings
from packy_agent.worker.loops.base.periodic import PeriodicLoop
from packy_agent.clients.packy_server import get_packy_server_client
from packy_agent.constants import HEARTBEAT_LOOP
from packy_agent.utils.network import get_actual_ip_address

logger = logging.getLogger(__name__)


def send_heartbeat():
    if not settings.is_worker_heartbeat_enabled():
        logger.debug('Heartbeat is disabled')
        return

    try:
        get_packy_server_client().update_status(version=packy_agent.__version__,
                                                ip_address=get_actual_ip_address())
    except Exception:
        logger.warning('Error while sending heartbeat', exc_info=True)


class HeartbeatLoop(PeriodicLoop):

    formal_name = HEARTBEAT_LOOP

    def __init__(self, period):
        super().__init__(period=period, callable_=send_heartbeat)



import time
import logging

from requests.exceptions import ConnectionError, HTTPError

from packy_agent.exceptions import ImproperlyConfiguredError, NoAccessTokenError
from packy_agent.configuration.sources.base import MappingBackedSource, READ_ONLY_EMPTY_DICT


logger = logging.getLogger(__name__)


SETTINGS_SUCCESS_CACHE_TIMEOUT_SECONDS = 600 / 100
SETTINGS_FAILURE_CACHE_TIMEOUT_SECONDS = 60 / 100


class Server(MappingBackedSource):

    def __init__(self, mapping=None, persist_callable=None, name=None):
        super().__init__(mapping, name=name)

        self.has_changed = False
        self._persist_callable = persist_callable
        self._mapping_expire_ts = 0

    def expire_cache(self):
        self._mapping_expire_ts = 0

    def ensure_ready(self):
        if self.is_ready or self.is_preparing:
            return

        self.get_mapping()

    def get_mapping(self):
        from packy_agent.clients.packy_server import get_packy_server_client

        if self._is_preparing:
            return super().get_mapping()

        self._is_preparing = True
        try:
            if time.time() >= self._mapping_expire_ts:
                try:
                    settings = get_packy_server_client().get_settings()
                    if self._persist_callable:
                        self._persist_callable(settings)
                except Exception as ex:
                    if isinstance(ex, ConnectionError):
                        logger.warning('Could not connect to server to get settings')
                    elif isinstance(ex, HTTPError):
                        logger.warning('Error while getting settings from server: %r', ex)
                    elif isinstance(ex, NoAccessTokenError):
                        logger.debug('No access token found. Probably agent is not activated yet')
                    elif isinstance(ex, ImproperlyConfiguredError):
                        logger.warning('Packy Server client was not configured: %s', ex)
                    else:
                        logger.exception('Error while getting settings from server')

                    self._mapping_expire_ts = time.time() + SETTINGS_FAILURE_CACHE_TIMEOUT_SECONDS
                else:
                    prev_mapping = super().get_mapping()
                    self.has_changed = (
                            prev_mapping is not READ_ONLY_EMPTY_DICT and settings != prev_mapping)
                    self.set_mapping(settings)  # To make it `ready`
                    self._mapping_expire_ts = time.time() + SETTINGS_SUCCESS_CACHE_TIMEOUT_SECONDS

            return super().get_mapping()
        finally:
            self._is_preparing = False



import logging

import cachetools

from packy_agent.utils.misc import safe_hexlify


PACKETS_CACHE_SIZE = 1000

logger = logging.getLogger(__name__)
icmp_responses = cachetools.TTLCache(PACKETS_CACHE_SIZE, ttl=60)


def buffer_icmp_response(data, receive_time):
    hexlified = safe_hexlify(data)
    logger.debug('Buffering ICMP response received at %.6f data: %s', receive_time, hexlified)

    if data in icmp_responses:
        logger.debug('ICMP response is already buffered at %.6f: %s',
                     icmp_responses.get(data), hexlified)
        return

    icmp_responses[data] = receive_time


def pop_buffered_icmp_response_receive_time(data):
    return icmp_responses.pop(data, None)


def is_non_empty_icmp_responses_buffer():
    return bool(icmp_responses)


def iter_icmp_responses():
    for data, receive_time in icmp_responses.items():
        yield data, receive_time


def remove_icmp_response_from_buffer(data):
    hexlified = safe_hexlify(data)
    logger.debug('REMOVING ICMP response: %s', hexlified)
    removed = icmp_responses.pop(data, None)
    if removed:
        logger.debug('REMOVED ICMP response: %s', hexlified)
    else:
        logger.debug('ICMP response was not found in buffer: %s', hexlified)



import logging

import flask

import packy_agent
from packy_agent.utils.pkg_resources import get_package_file_full_path
from packy_agent.console.views.index import IndexView
from packy_agent.console.views.login import LoginView, LoginFailureView
from packy_agent.console.views.logout import LogoutView
from packy_agent.console.views.success import SuccessView
from packy_agent.console.views.failure import FailureView
from packy_agent.console.views.network import NetworkView
from packy_agent.console.views.control import ControlView
from packy_agent.console.views.upgrade import UpgradeView
from packy_agent.console.views.reset import ResetView
from packy_agent.console.views.debug import DebugView
from packy_agent.console.views.activate import ActivateView, ActivationFailureView
from packy_agent.console.views import errors as error_handlers
from packy_agent.utils.logging import configure_logging, configure_logging_basic
from packy_agent.utils.installation import (
    dump_version, remove_upgrade_in_progress_lock,
    is_upgrade_in_progress)
from packy_agent.domain_logic.managers.network import network_manager
from packy_agent.utils.platforms import is_inside_docker_container
from packy_agent.clients.sentry import init_sentry_client
from packy_agent.configuration.settings import settings
from packy_agent.utils.misc import generate_flask_secret_key
from packy_agent.domain_logic.constants import CONSOLE_COMPONENT


def configure(app):
    # TODO(dmu) LOW: Refactor to make autooverride from options if there are more options to
    #                override
    mapping = settings.get_console_flask()
    if 'SECRET_KEY' not in mapping:
        settings.set_console_flask_secret_key(generate_flask_secret_key())
        mapping = settings.get_console_flask()

    app.config.from_mapping(mapping)


def setup_routes(app):
    app.add_url_rule('/', view_func=IndexView.as_view('index'))
    app.add_url_rule('/network/', view_func=NetworkView.as_view('network'))
    app.add_url_rule('/control/', view_func=ControlView.as_view('control'))
    app.add_url_rule('/upgrade/', view_func=UpgradeView.as_view('upgrade'))
    app.add_url_rule('/reset/', view_func=ResetView.as_view('reset'))
    app.add_url_rule('/debug/', view_func=DebugView.as_view('debug'))
    app.add_url_rule('/success/', view_func=SuccessView.as_view('success'))
    app.add_url_rule('/failure/', view_func=FailureView.as_view('failure'))
    app.add_url_rule('/activate/', view_func=ActivateView.as_view('activate'))
    app.add_url_rule('/activation-failure/',
                     view_func=ActivationFailureView.as_view('activation_failure'))
    app.add_url_rule('/login/', view_func=LoginView.as_view('login'))
    app.add_url_rule('/login-failure/', view_func=LoginFailureView.as_view('login_failure'))
    app.add_url_rule('/logout/', view_func=LogoutView.as_view('logout'))


def setup_context_processors(app):

    @app.context_processor
    def extra_context_processor():
        return {
            'version': packy_agent.__version__,
            'is_inside_docker_container': is_inside_docker_container(),
            'is_network_configuration_supported': network_manager.is_network_configuration_supported(),
            'is_upgrade_in_progress': is_upgrade_in_progress,
            'settings': settings,
        }


def setup_error_handlers(app):
    app.register_error_handler(500, error_handlers.handle_http500)
    app.register_error_handler(401, error_handlers.handle_http401)


def get_app(args=None):
    settings.set_runtime('component', CONSOLE_COMPONENT)

    if args:
        log_level = args.log_level
        cargs_console = vars(args)
        del cargs_console['log_level']
        cargs = {
            'log_level': log_level,
            'console': cargs_console,
        }
        settings.set_commandline_arguments(cargs)
    else:
        log_level = logging.WARNING

    configure_logging_basic(log_level)

    settings.validate()

    init_sentry_client()
    configure_logging(log_level)

    app = flask.Flask(__name__,
                      template_folder=get_package_file_full_path(__name__, 'templates'),
                      static_folder=get_package_file_full_path(__name__, 'static'),
                      static_url_path='/assets')
    configure(app)
    setup_routes(app)
    setup_context_processors(app)
    setup_error_handlers(app)

    remove_upgrade_in_progress_lock()
    dump_version(settings.get_console_version_dump_filename(),
                 settings.get_console_version_dump_variable_name())
    settings.update_started_at_ts()

    return app



import gevent
from gevent.queue import Queue

from packy_agent.worker.loops.base.consumer import ConsumerLoop


def test_scheduled_loop_survives_exception_in_callable():
    queue = Queue()
    counter = 0

    def call_me(item):
        nonlocal counter
        if not item:
            raise Exception('Test exception')
        counter += 1

    loop = ConsumerLoop(queue, call_me)
    loop.start()
    queue.put(1)
    gevent.sleep(1)
    counter_value_before_exception = counter
    assert counter_value_before_exception
    queue.put(None)
    gevent.sleep(1)
    queue.put(1)
    gevent.sleep(1)
    assert counter > counter_value_before_exception
    loop.stop()
    gevent.sleep(1)



import argparse

from packy_agent.configuration.settings import settings


def get_base_argument_parser(module=None, kwargs=None, default_log_level=None):
    kwargs = kwargs or {}
    if module:
        kwargs['prog'] = 'python -m {}'.format(module)
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                     **kwargs)
    parser.add_argument(
        '--log-level', help='Log level',
        default=settings.get_log_level() if default_log_level is None else default_log_level)

    return parser



from urllib.parse import urljoin

import flask
from flask.views import MethodView

from packy_agent.console.forms.network import NetworkForm
from packy_agent.configuration.settings import settings
from packy_agent.domain_logic.managers.network import network_manager
from packy_agent.domain_logic.managers.reset import reset_manager
from packy_agent.console.views.base import smart_redirect
from packy_agent.utils.auth import activation_and_authentication_required
from packy_agent.console.forms.action import ActionForm


DELAY_SECONDS = 5


class NetworkView(MethodView):

    def get_form_context(self):
        is_network_configuration_supported = network_manager.is_network_configuration_supported()
        network_interface = network_manager.get_configurable_network_interface()

        if is_network_configuration_supported and network_interface:
            form_kwargs = network_manager.get_configuration(network_interface)
            form_kwargs['network_interface'] = network_interface
            form = NetworkForm(**form_kwargs)
        else:
            form = None

        context = {
            'form': form,
            'is_network_configuration_enabled': settings.is_network_configuration_enabled(),
            'network_interface': network_interface,
            'active_menu_item': 'network',
        }
        return context

    @activation_and_authentication_required
    def get(self):
        context = self.get_form_context()
        return flask.render_template('network.html', **context)

    @activation_and_authentication_required
    def post(self):
        action_form = ActionForm()
        if action_form.validate():
            action = action_form.action.data
            if action == 'network_reset':
                if settings.is_network_configuration_enabled():
                    reset_manager.reset_network_configuration(reboot_delay_seconds=DELAY_SECONDS)

                    flask.flash(f'Reboot will start in {DELAY_SECONDS} seconds...')
                    return smart_redirect('success', 'index', button_text='Continue')
                else:
                    flask.flash('Network configuration is not allowed.')
                    return smart_redirect('success', 'network')

        form = NetworkForm()
        if form.validate():

            if form.dhcp.data:
                network_manager.set_dhcp(interface=form.network_interface.data,
                                         reboot_delay_seconds=DELAY_SECONDS)
                back_endpoint = 'network'
            else:
                ip_address = form.ip_address.data
                kwargs = dict(
                    ip_address=ip_address,
                    subnet_mask=form.subnet_mask.data,
                    default_gateway=form.default_gateway.data,
                    interface=form.network_interface.data,
                    reboot_delay_seconds=DELAY_SECONDS,
                )
                if form.name_servers.data:
                    kwargs['name_servers'] = form.name_servers.data.split(',')

                network_manager.set_static_ip_address(**kwargs)

                base_url = settings.get_console_base_url(ip_address)
                back_endpoint = urljoin(base_url, flask.url_for('network'))

            if settings.is_network_configuration_enabled():
                flask.flash(f'Changes will take affect after reboot in {DELAY_SECONDS} seconds...')
            else:
                flask.flash('Dry run. Please, see logs for more information')
                back_endpoint = 'network'  # no changes, so...

            return smart_redirect('success', back_endpoint)

        context = self.get_form_context()
        context['form'] = form
        return flask.render_template('network.html', **context)



import logging

from ryu.lib.packet.icmp import icmp
from ryu.lib.packet.ipv4 import ipv4
from ryu.lib.packet.packet import Packet

from packy_agent.utils.misc import safe_hexlify


logger = logging.getLogger(__name__)


def retrieve_ip_and_icmp_packets(data):
    packet = Packet(data, parse_cls=ipv4)

    protocol_0 = packet.protocols[0]
    if isinstance(protocol_0, ipv4):  # is IP packet?
        ip_packet = protocol_0
        logger.debug('Processing IPv4 packet checksum: %s, from: %s', ip_packet.csum, ip_packet.src)
    else:
        logger.debug('Data is not an IPv4 packet: %s', safe_hexlify(data))
        return

    protocol_1 = packet.protocols[1]
    if isinstance(protocol_1, icmp):
        icmp_packet = protocol_1
        logger.debug('Processing ICMP packet IP checksum: %s, from: %s, type: %s, code: %s',
                     ip_packet.csum, ip_packet.src, icmp_packet.type, icmp_packet.code)
    else:
        logger.debug('Data does not contain an ICMP packet: %s', safe_hexlify(data))
        return

    return ip_packet, icmp_packet



import logging
import sys

from tabulate import tabulate

from packy_agent.configuration.settings import settings
from packy_agent.utils.cli import get_base_argument_parser
from packy_agent.utils.logging import configure_logging, configure_logging_basic
from packy_agent.domain_logic.modules.traceroute import base
from packy_agent.domain_logic.modules.traceroute.constants import TRACEROUTE_METHODS, UDP_METHOD


def do_traceroute(host, timeout=3, probe_count=1, packet_size_bytes=60, max_hops=30,
                  method=UDP_METHOD, max_parallelism=1):
    agg_results = base.traceroute(host, timeout=timeout, probe_count=probe_count,
                                  packet_size_bytes=packet_size_bytes, max_hops=max_hops,
                                  method=method, max_parallelism=max_parallelism)
    if not agg_results:
        print('No results')
        return

    headers = ['#', 'target', 'average']
    for agg_result in agg_results:
        headers.extend(agg_result.keys() - set(headers) - {'host'})

    rows = [
        [x, result.get('host')] + [result.get(k) for k in headers[2:]] for x, result in
        enumerate(agg_results, start=1)]

    print(tabulate(rows, headers=headers, tablefmt='orgtbl'))


def entry():
    configure_logging_basic(logging.WARNING)
    parser = get_base_argument_parser(__loader__.name)

    parser.add_argument('host')
    parser.add_argument('--max-hops', help='max hops', type=int, default=30)
    parser.add_argument('--probe-count', help='probe count', type=int, default=3)
    parser.add_argument('--packet-size-bytes', help='IP packet size in bytes',
                        type=int, default=60)
    parser.add_argument('--max-parallelism', help='max parallelism', type=int, default=1)
    parser.add_argument('--timeout', help='timeout', type=float, default=3)
    parser.add_argument('-M', '--method', help='method', choices=TRACEROUTE_METHODS,
                        default=UDP_METHOD)

    args = parser.parse_args()
    settings.set_commandline_arguments(vars(args))
    configure_logging(logging.WARNING)

    return do_traceroute(args.host, timeout=args.timeout, probe_count=args.probe_count,
                         packet_size_bytes=args.packet_size_bytes,
                         max_hops=args.max_hops, method=args.method,
                         max_parallelism=args.max_parallelism)


if __name__ == '__main__':
    sys.exit(entry())



from unittest.mock import patch

import gevent
from sqlalchemy import create_engine

from packy_agent.configuration.settings import settings
from packy_agent.domain_logic.models.sqlalchemy.base import Session
from packy_agent.domain_logic.models.sqlalchemy.measurement import Measurement
from packy_agent.tests.worker.loops.main.base import start_main_loop, stop_main_loop

TASKS = [
    {
        'id': 1,
        'module_name': 'ping',
        'args': [],
        'kwargs': {
            'number_of_pings': 3,
            'host': '8.8.8.8',
            'interval_ms': 0,
            'packet_size': 64
        },
        'schedule': '*/2 * * * * *'
    },
]


def test_purger():
    main_loop = start_main_loop(TASKS)
    gevent.sleep(1)
    main_loop.task_results_consumer.stop()
    main_loop.task_results_submitter.stop()
    while main_loop.task_results_consumer.is_running or main_loop.task_results_submitter.is_running:
        gevent.sleep(1)

    gevent.sleep(5)  # Let get measurements
    engine = create_engine(settings.get_database_url())
    with Session(engine) as session:
        measurements = session.query(Measurement).order_by(Measurement.created_at_ts).all()

    assert measurements

    with patch('packy_agent.worker.loops.task_results.purger.get_available_disk_space',
               new=lambda path=None: 0):
        gevent.sleep(2)

    with Session(engine) as session:
        measurements = session.query(Measurement).order_by(Measurement.created_at_ts).all()

    assert not measurements

    stop_main_loop(main_loop)



import time
import re
import logging
from functools import partial
from collections import OrderedDict
import threading
from collections import Mapping

from packy_agent.exceptions import ImproperlyConfiguredError
from packy_agent.utils.misc import SENTINEL, yaml_coerce
from packy_agent.configuration.sources.base import MappingBackedSource, MutableMappingBackedSource
from packy_agent.configuration.sources.key_value_storage import KeyValueStorageBackedSource
from packy_agent.configuration.sources.defaults import ConfigurationDefaults
from packy_agent.configuration.sources.environment import Environment
from packy_agent.configuration.sources.file import File as FileSource
from packy_agent.configuration.sources.server import Server as ServerSource
from packy_agent.utils.collections import deep_update
from packy_agent.configuration.base import SUPERVISOR, CONSOLE, WORKER, WATCHDOG


TABLENAME = 'settings'
GETTER_PREFIX = 'get_'
IS_PREFIX = 'is_'
SETTER_PREFIX = 'set_'

METHOD_PATTERN_RE = re.compile(
    f'^(?P<type>{GETTER_PREFIX}|{IS_PREFIX}|{SETTER_PREFIX})'
    f'(?:(?P<section>{WORKER}|{CONSOLE}|{WATCHDOG}|{SUPERVISOR})_)?(?P<key>.*)?')


logger = logging.getLogger(__name__)
threading_local = threading.local()


# TODO(dmu) MEDIUM: Implement ConfigurationSourceBase interface?
class Settings:

    def __init__(self):
        # Prefilled with 'invalid_access_tokens' for optimization purposes
        self._runtime = MutableMappingBackedSource({'invalid_access_tokens': set()},
                                                   name='Local runtime')

        self._key_value_storage = None
        self._commandline_arguments = {}

        self._cached_settings = KeyValueStorageBackedSource(
            self.get_cached_settings_tablename, self.get_key_value_storage_filename,
            name='Cached'
        )
        self._server_source = ServerSource(
            persist_callable=lambda x: self._cached_settings.replace(x),
            name='From server',
        )

        self._local_settings = KeyValueStorageBackedSource(
            self.get_local_settings_tablename, self.get_key_value_storage_filename,
            name='Local',
        )

        self._sources = [
            self._runtime,
            MappingBackedSource(self._commandline_arguments, name='From command line'),
            self._server_source,
            self._cached_settings,
            self._local_settings,
            Environment(name='From environment variables'),
            FileSource(self.get_settings_filename, name='Settings file'),
            ConfigurationDefaults(),
        ]

    def __getattr__(self, item):
        if isinstance(item, str):
            match = METHOD_PATTERN_RE.match(item)
            if match:
                type_ = match.group('type')
                section = match.group('section')
                key = match.group('key')
                if type_ in (GETTER_PREFIX, IS_PREFIX):
                    if section:
                        return partial(self.get_subkey, section, key)
                    else:
                        return partial(self.get, key)
                elif type_ == SETTER_PREFIX:
                    if section:
                        return partial(self.set_subkey, section, key)
                    else:
                        return partial(self.set, key)
                else:
                    raise ValueError(f'Unknown prefix {type_}')  # should never get here

        raise AttributeError("{!r} object has no attribute {!r}".format(self.__class__, item))

    def labeled_items(self):
        for source in self._sources:
            yield source.name, source.items() if source.is_ready else ()

    def items(self):
        items = OrderedDict()
        for source in reversed(self._sources):
            items.update(source.items())

        return items

    def get(self, key, default=None):
        sources = self._sources
        for source in sources:
            source.ensure_ready()

        sources_iterator = iter(sources)
        for source in sources_iterator:
            if not source.is_ready:
                continue

            value = source.get(key, SENTINEL)
            if value is not SENTINEL:
                if isinstance(value, Mapping):
                    # This part is about deep update (for logging configuration mostly)
                    last_value = value
                    new_value = {}
                    for deeper_source in reversed(list(sources_iterator)):
                        if deeper_source.is_ready:
                            value = deeper_source.get(key, SENTINEL)
                            if value is not SENTINEL:
                                deep_update(new_value, value)

                    deep_update(new_value, last_value)
                    value = new_value

                break
        else:
            value = default

        return value

    def get_subkey(self, key, subkey, default=None):
        return (self.get(key) or {}).get(subkey, default)

    def set(self, key, value):
        coerced_value = yaml_coerce(value)
        try:
            self._local_settings[key] = coerced_value
        except TypeError:
            raise ImproperlyConfiguredError(
                'Unable to set {}={}: local settings key value storage is '
                'not configured'.format(key, value))

    def set_subkey(self, key, subkey, value):
        super_value = self._local_settings.get(key)
        super_value = super_value.copy() if super_value else {}
        super_value[subkey] = value
        self.set(key, super_value)

    def set_runtime(self, key, value):
        self._runtime[key] = value

    def delete(self, key):
        try:
            del self._local_settings[key]
        except TypeError:
            raise ImproperlyConfiguredError(
                "Unable to delete '{}': local settings key value storage is "
                "not configured".format(key))

    def clear(self):
        self._local_settings.clear()

    def set_commandline_arguments(self, arguments):
        self._commandline_arguments.clear()
        self._commandline_arguments.update(arguments)

    def validate(self):
        if not self.get_server_base_url():
            raise ImproperlyConfiguredError("'server_base_url' must be explicitly set")
        if not self.get_component():
            raise ImproperlyConfiguredError("'component' must be explicitly set")

    def activate(self, client_id, access_token, refresh_token, agent_key, agent_name):
        self.set('client_id', client_id)  # needed for refresh token only
        self.set('access_token', access_token)
        self.set('refresh_token', refresh_token)
        self.set('agent_key', agent_key)
        self.set('agent_name', agent_name)
        self.set('activated', True)

        self._server_source.expire_cache()

    def deactivate(self):
        self.delete('client_id')
        self.delete('access_token')
        self.delete('refresh_token')
        self.delete('agent_key')
        self.delete('agent_name')
        self.set('activated', False)

    def has_settings_changed_on_server(self):
        return self._server_source.has_changed

    def update_started_at_ts(self, component=None):
        component = component or self.get_component()
        if not component:
            raise ImproperlyConfiguredError('Component must be provided')

        self.set_subkey(component, 'started_at_ts', time.time())

    def update_restarted_at_ts(self, component):
        self.set_subkey(component, 'restarted_at_ts', time.time())

    def update_rebooted_at_ts(self, delta=0):
        self.set('rebooted_at_ts', time.time() + delta)

    #######################################
    # Explicit settings getters and setters

    def get_log_level(self):
        return self.get('log_level')

    def get_settings_filename(self):
        return self.get('settings_filename')

    def get_database_filename(self):
        return self.get('database_filename')

    def get_key_value_storage_filename(self):
        return self.get_database_filename()

    def get_database_url(self):
        database_filename = self.get_database_filename()
        if database_filename:
            return 'sqlite:///' + database_filename

    def get_cached_settings_tablename(self):
        return self.get('cached_settings_tablename')

    def get_local_settings_tablename(self):
        return self.get('local_settings_tablename')

    def get_server_base_url(self):
        server_base_url = self.get('server_base_url')
        if server_base_url:
            return server_base_url.rstrip('/') + '/'

    def get_console_base_url(self, ip_address=None):
        from packy_agent.utils.network import get_actual_ip_address
        # TODO(dmu) LOW: Unhardcode schema: `http://`
        ip_address = ip_address or get_actual_ip_address()
        url = 'http://' + ip_address
        port = self.get_console_http_port()
        if port != 80:
            url += f':{port}'

        return url + '/'

    def set_console_flask_secret_key(self, key):
        mapping = self.get_console_flask()
        mapping = mapping.copy() if mapping else {}
        mapping['SECRET_KEY'] = key
        self.set_console_flask(mapping)

    def reset_network_data_usage(self):
        self.delete('bytes_sent')
        self.delete('bytes_received')


# We need all greenlets to share the same settings object to share runtime-settings
# (stored in RAM only) and to avoid excess access to externally stored settings
# (i.e. on Packy Server)
settings = Settings()



from collections import Mapping, MutableMapping


def deep_update(base_dict, update_with):
    for key, value in update_with.items():
        if isinstance(value, Mapping):
            base_dict_value = base_dict.get(key)
            if isinstance(base_dict_value, MutableMapping):
                deep_update(base_dict_value, value)
            else:
                base_dict[key] = value
        else:
            base_dict[key] = value


def set_if(dict_, key, value, if_callable=lambda x: x):
    if if_callable(value):
        dict_[key] = value



from unittest.mock import patch

from packy_agent.domain_logic.managers.control import control_manager


def test_get_pid():
    with patch('packy_agent.domain_logic.managers.control.run_supervisor_command') as mock:
        mock.return_value = 'worker RUNNING pid 11, uptime 1 minute'
        assert control_manager.get_pid('worker') == 11
        mock.assert_called()


def test_get_status():
    with patch('packy_agent.domain_logic.managers.control.run_supervisor_command') as mock:
        mock.return_value = 'worker RUNNING pid 11, uptime 1 minute'
        assert control_manager.get_status('worker') == ('RUNNING', 11, '1 minute')
        mock.assert_called()


def test_get_nginx_status():
    with patch('packy_agent.domain_logic.managers.control.run_supervisor_command') as sup_mock:
        with patch('packy_agent.domain_logic.managers.control.nginx_service') as nginx_mock:
            with patch('packy_agent.domain_logic.managers.control.'
                       'is_inside_docker_container') as is_inside_docker_container_mock:
                sup_mock.return_value = 'nginx STOPPED'
                is_inside_docker_container_mock.return_value = False
                nginx_mock.is_running.return_value = True
                nginx_mock.get_pid.return_value = 12
                assert control_manager.get_status('nginx') == ('RUNNING', 12, None)
                sup_mock.assert_not_called()


def test_get_nginx_status_in_docker():
    with patch('packy_agent.domain_logic.managers.control.run_supervisor_command') as sup_mock:
        with patch('packy_agent.domain_logic.managers.control.nginx_service') as nginx_mock:
            with patch('packy_agent.domain_logic.managers.control.'
                       'is_inside_docker_container') as is_inside_docker_container_mock:
                sup_mock.return_value = 'nginx RUNNING pid 11, uptime 1 minute'
                is_inside_docker_container_mock.return_value = True
                nginx_mock.is_running.return_value = False
                nginx_mock.get_pid.return_value = None
                assert control_manager.get_status('nginx') == ('RUNNING', 11, '1 minute')
                sup_mock.assert_called()


def test_get_worker_status_line():
    with patch('packy_agent.domain_logic.managers.control.run_supervisor_command') as mock:
        mock.return_value = 'worker RUNNING pid 11, uptime 1 minute'
        assert control_manager.get_worker_status_line() == 'RUNNING (uptime 1 minute)'
        mock.assert_called()


def test_is_running():
    with patch('packy_agent.domain_logic.managers.control.run_supervisor_command') as mock:
        mock.return_value = 'worker RUNNING pid 11, uptime 1 minute'
        assert control_manager.is_running('worker')
        mock.assert_called()


def test_is_stopped():
    with patch('packy_agent.domain_logic.managers.control.run_supervisor_command') as mock:
        mock.return_value = 'worker STOPPED pid 11, uptime 1 minute'
        assert control_manager.is_stopped('worker')
        mock.assert_called()



import os
import stat
from io import StringIO

from packy_agent.utils.services.base import ServiceBase
from packy_agent.utils.shell import run_command
from packy_agent.utils.output import write_to_console_or_file


class SystemdService(ServiceBase):

    RELOAD_SYSTEMD_CONFIGURATION_COMMAND = 'systemctl daemon-reload'
    MAKE_START_ON_REBOOT_COMMAND_TEMPLATE = 'update-rc.d {name} defaults'

    SERVICE_COMMAND_TEMPLATE = 'service {name} {command}'
    SYSTEMCTL_COMMAND_TEMPLATE = 'systemctl {command} {name}.service'

    INITD_PATH_TEMPLATE = '/etc/init.d/{name}'
    # TODO(dmu) MEDIUM: Is `UNIT_SERVICE_PATH_TEMPLATE` different on OSes other than Ubuntu
    UNIT_SERVICE_PATH_TEMPLATE = '/lib/systemd/system/{name}.service'

    def __init__(self, name):
        super(SystemdService, self).__init__(name)

        self.make_start_on_reboot_command = self.MAKE_START_ON_REBOOT_COMMAND_TEMPLATE.format(
            name=name)

        self.initd_path = self.INITD_PATH_TEMPLATE.format(name=name)
        self.unit_service_path = self.UNIT_SERVICE_PATH_TEMPLATE.format(name=name)

    @classmethod
    def get_template(cls, is_systemctl=False):
        return cls.SYSTEMCTL_COMMAND_TEMPLATE if is_systemctl else cls.SERVICE_COMMAND_TEMPLATE

    @classmethod
    def reload_systemd_configuration(cls, raise_exception=False):
        run_command(cls.RELOAD_SYSTEMD_CONFIGURATION_COMMAND, raise_exception=raise_exception)

    def run_basic_command(self, command, is_systemctl=False, raise_exception=False,
                          is_async=False):
        command = self.get_template(
            is_systemctl=is_systemctl).format(name=self.name, command=command)
        return run_command(command, raise_exception=raise_exception, is_async=is_async)

    def start(self, is_systemctl=False, raise_exception=False):
        self.run_basic_command('start', is_systemctl=is_systemctl, raise_exception=raise_exception)

    def stop(self, is_systemctl=False, raise_exception=False, is_async=False):
        self.run_basic_command('stop', is_systemctl=is_systemctl, raise_exception=raise_exception,
                               is_async=is_async)

    def restart(self, is_systemctl=False, raise_exception=False):
        self.run_basic_command('restart', is_systemctl=is_systemctl,
                               raise_exception=raise_exception)

    def reload(self, is_systemctl=False, raise_exception=False):
        self.run_basic_command('reload', is_systemctl=is_systemctl,
                               raise_exception=raise_exception)

    def make_start_on_reboot(self, raise_exception=False):
        run_command(self.make_start_on_reboot_command, raise_exception=raise_exception)

    def enable_unit_service(self, raise_exception=False):
        self.run_basic_command('enable', is_systemctl=True, raise_exception=raise_exception)

    def update_initd_script(self, content):
        initd_path = self.initd_path
        write_to_console_or_file(initd_path, content)
        os.chmod(initd_path, os.stat(initd_path).st_mode | stat.S_IEXEC)
        self.reload_systemd_configuration(raise_exception=True)
        self.make_start_on_reboot(raise_exception=True)

    def update_unit_service_configuration(self, content):
        write_to_console_or_file(self.unit_service_path, content)
        self.enable_unit_service(raise_exception=True)
        self.reload_systemd_configuration(raise_exception=True)
        self.restart(is_systemctl=True, raise_exception=True)

    def is_active(self, raise_exception=False):
        output = self.run_basic_command('is-active',
                                        is_systemctl=True, raise_exception=raise_exception)
        return output == 'active'

    def is_running(self):
        return self.is_active()

    def get_property(self, name, raise_exception=False):
        prefix = name + '='
        output = self.run_basic_command('show', is_systemctl=True, raise_exception=raise_exception)
        if output:
            for line in StringIO(output.decode('utf-8')):
                line = line.rstrip('\n\r')
                if line.startswith(prefix):
                    return line[len(prefix):]

    def get_pid(self, raise_exception=False):
        value = self.get_property('MainPID', raise_exception=raise_exception)
        try:
            return int(value)
        except (TypeError, ValueError):
            return None


packy_service = SystemdService('packy')
nginx_service = SystemdService('nginx')
log2ram_service = SystemdService('log2ram')



from unittest.mock import MagicMock, patch

import pytest

from packy_agent.watchdog.service import WatchdogService


@pytest.mark.usefixtures('activated', 'watchdog_started_long_ago')
def test_program_is_running():
    with patch('packy_agent.watchdog.service.control_manager.is_running') as is_running:
        is_running.return_value = True
        service = WatchdogService()

        service.reboot = MagicMock()
        service.restart_worker = MagicMock()
        service.restart_console = MagicMock()
        service.start_worker = MagicMock()

        service.resolve_online_status = MagicMock(return_value=None)
        service.resolve_worker_process_status = MagicMock(return_value=None)
        service.resolve_worker_response = MagicMock(return_value=None)

        service.loop_iteration()

        service.start_worker.assert_not_called()
        service.restart_worker.assert_not_called()
        service.reboot.assert_not_called()


@pytest.mark.usefixtures('activated', 'watchdog_started_long_ago')
def test_program_is_not_running():
    with patch('packy_agent.watchdog.service.control_manager.is_running') as is_running:
        is_running.return_value = False
        service = WatchdogService()

        service.reboot = MagicMock()
        service.restart_worker = MagicMock()
        service.restart_console = MagicMock()
        service.start_worker = MagicMock()

        service.resolve_online_status = MagicMock(return_value=None)
        service.resolve_worker_process_status = MagicMock(return_value=None)
        service.resolve_worker_response = MagicMock(return_value=None)

        service.loop_iteration()

        service.start_worker.assert_called()
        service.restart_worker.assert_not_called()
        service.reboot.assert_not_called()



import os
from unittest.mock import MagicMock, patch

import pytest

from packy_agent.watchdog.service import WatchdogService


def get_pid_mock(target):
    if target == 'worker':
        return os.getpid()

    return 9999999999999999999999


@pytest.mark.usefixtures('activated', 'watchdog_started_long_ago')
def test_process_is_running():
    with patch('packy_agent.watchdog.service.control_manager.get_pid', get_pid_mock) as get_pid:
        service = WatchdogService()

        service.reboot = MagicMock()
        service.restart_worker = MagicMock()
        service.start_worker = MagicMock()

        service.resolve_online_status = MagicMock(return_value=None)
        service.resolve_worker_supervisor_program_status = MagicMock(return_value=None)
        service.resolve_worker_response = MagicMock(return_value=None)

        service.loop_iteration()

        service.start_worker.assert_not_called()
        service.restart_worker.assert_not_called()
        service.reboot.assert_not_called()


@pytest.mark.usefixtures('activated', 'watchdog_started_long_ago')
def test_process_is_not_running():
    with patch('packy_agent.watchdog.service.control_manager.get_pid') as get_pid:
        get_pid.return_value = 9999999999999999999999
        service = WatchdogService()

        service.reboot = MagicMock()
        service.restart_worker = MagicMock()
        service.start_worker = MagicMock()

        service.resolve_online_status = MagicMock(return_value=None)
        service.resolve_worker_supervisor_program_status = MagicMock(return_value=None)
        service.resolve_worker_response = MagicMock(return_value=None)

        service.loop_iteration()

        service.start_worker.assert_not_called()
        service.restart_worker.assert_called()
        service.reboot.assert_not_called()



import logging
from tailer import Tailer

import flask
from flask.views import MethodView

from packy_agent.configuration.settings import settings
from packy_agent.utils.auth import activation_and_authentication_required

TAIL_SIZE = 100


logger = logging.getLogger(__name__)


class CustomTailer(Tailer):
    def splitlines(self, data):
        return super().splitlines(data.decode('utf-8', errors='ignore'))


def tail(filename, size=TAIL_SIZE):
    try:
        with open(filename, 'rb') as f:
            return '\n'.join(CustomTailer(f).tail(size))
    except Exception as ex:
        return f'Exception occurred while reading log file: {ex!r}'


class DebugView(MethodView):

    @activation_and_authentication_required
    def get(self):

        logs = (
            ('Supervisor log', settings.get_supervisor_log_filename(),
             tail(settings.get_supervisor_log_filename())),
            ('Worker stdout log', settings.get_worker_stdout_log_filename(),
             tail(settings.get_worker_stdout_log_filename())),
            ('Worker stderr log', settings.get_worker_stderr_log_filename(),
             tail(settings.get_worker_stderr_log_filename())),
            ('Console stdout log', settings.get_console_stdout_log_filename(),
             tail(settings.get_console_stdout_log_filename())),
            ('Console stderr log', settings.get_console_stderr_log_filename(),
             tail(settings.get_console_stderr_log_filename())),
        )

        context = {
            'enumerated_settings': enumerate(settings.labeled_items(), start=1),
            'logs': logs,
            'active_menu_item': 'debug',
        }

        return flask.render_template('debug.html', **context)



import tempfile

import pytest

from packy_agent.configuration.sources.base import YamlFileBackedSource


def test_yaml_file_backed_source():
    with tempfile.NamedTemporaryFile(buffering=0) as f:
        f.write(b'value1: 10\nvalue2: 15')

        source = YamlFileBackedSource(f.name)
        assert set(source) == {'value1', 'value2'}
        assert source['value1'] == 10
        assert source['value2'] == 15
        assert 'value1' in source
        assert source.get('value1') == 10
        assert source.get('value3', 4) == 4
        assert sorted(source.keys()) == ['value1', 'value2']
        assert sorted(source.values()) == [10, 15]
        assert sorted(source.items()) == [('value1', 10), ('value2', 15)]

        with pytest.raises(TypeError) as ex:
            source['value1'] = 0

        assert "'YamlFileBackedSource' object does not support item assignment" in str(ex.value)

        with pytest.raises(TypeError) as ex:
            del source['value1']
        assert "'YamlFileBackedSource' object does not support item deletion" in str(ex.value)

        with pytest.raises(AttributeError) as ex:
            source.pop('value1')
        assert "'YamlFileBackedSource' object has no attribute 'pop'" in str(ex.value)



import logging

from packy_agent.constants import HTTP_MODULE_LOOP
from packy_agent.worker.loops.modules.base import ScheduledProducerTaskLoop
from packy_agent.domain_logic.modules.http.base import get_http_measurement

MBIT = 1000000.0


logger = logging.getLogger(__name__)


def http_impl(*args, **kwargs):
    url = kwargs.get('url')
    if not url:
        logger.debug('URLs are not configured for `http` module')
        return

    follow_redirects = kwargs.get('follow_redirects', False)

    # TODO(dmu) MEDIUM: Figure out what happens inside `curl` library and remove this
    #                   workaround
    for _ in range(10):
        measurement = get_http_measurement(url, follow_redirects=follow_redirects)
        if not measurement.is_success or (measurement.total_ms < 4000000 and
                                          measurement.namelookup_ms < 4000000):
            break  # we got the measurement
    else:
        measurement = None
        logger.error('Failed to get proper measurement from curl for %s', url)

    return measurement


class HTTPTaskLoop(ScheduledProducerTaskLoop):

    formal_name = HTTP_MODULE_LOOP

    def __init__(self, schedule, args=(), kwargs=None, outbound_queue=None):
        super().__init__(schedule=schedule, callable_=http_impl, args=args, kwargs=kwargs,
                         outbound_queue=outbound_queue)



from packy_agent.tests.fixtures import *


def pytest_addoption(parser):
    parser.addoption('--run-slow', action='store_true', default=False, help='run slow tests')


def pytest_collection_modifyitems(config, items):
    if config.getoption('--run-slow'):
        # --run-slow given in cli: do not skip slow tests
        return

    skip_slow = pytest.mark.skip(reason='need --run-slow option to run')
    for item in items:
        if 'slow' in item.keywords:
            item.add_marker(skip_slow)



import time

from unittest.mock import patch, call, MagicMock

from packy_agent.tests.fixtures.configuration import override_settings
from packy_agent.domain_logic.managers.control import control_manager


def test_stop():
    with patch('packy_agent.domain_logic.managers.control.run_supervisor_command') as run_mock:
        with patch('packy_agent.domain_logic.managers.control.Process') as process_cls_mock:
            with patch('packy_agent.domain_logic.managers.control.'
                       'control_manager.terminate_process'):
                with patch('packy_agent.domain_logic.managers.control.'
                           'control_manager.kill_process'):
                    process_mock = MagicMock()
                    process_mock.is_running.return_value = False
                    process_cls_mock.return_value = process_mock

                    def side_effect(command, cargs, *args, **kwargs):
                        return (f'{cargs[0]} RUNNING pid 11, uptime 1 minute'
                                if command == 'status' else None)

                    run_mock.side_effect = side_effect
                    control_manager.stop('worker')
                    run_mock.assert_has_calls(
                        (call('stop', ('worker',), is_async=False,
                              async_delay_seconds=None, async_dev_null=False),), any_order=True)
                    control_manager.terminate_process.assert_not_called()
                    control_manager.kill_process.assert_not_called()


def test_terminate():
    with patch('packy_agent.domain_logic.managers.control.run_supervisor_command') as run_mock:
        with patch('packy_agent.domain_logic.managers.control.Process') as process_cls_mock:
            with patch('packy_agent.domain_logic.managers.control.'
                       'control_manager.terminate_process'):
                with patch('packy_agent.domain_logic.managers.control.'
                           'control_manager.kill_process'):
                    with override_settings({'process_stop_timeout_seconds': 0.2,
                                            'process_status_check_period_seconds': 0.02}):

                        terminate_at = time.time() + 0.5

                        def process_is_running_side_effect():
                            if time.time() >= terminate_at:
                                return False

                            return True

                        process_mock = MagicMock()
                        process_mock.is_running.side_effect = process_is_running_side_effect
                        process_cls_mock.return_value = process_mock

                        def side_effect(command, cargs, *args, **kwargs):
                            if command == 'status':
                                if time.time() >= terminate_at:
                                    return f'{cargs[0]} STOPPED'
                                else:
                                    return f'{cargs[0]} RUNNING pid 11, uptime 1 minute'

                        run_mock.side_effect = side_effect
                        control_manager.stop('worker')
                        run_mock.assert_has_calls(
                            (call('stop', ('worker',), is_async=False,
                                  async_delay_seconds=None, async_dev_null=False),), any_order=True)
                        control_manager.terminate_process.assert_called()
                        control_manager.kill_process.assert_not_called()


def test_kill():
    with patch('packy_agent.domain_logic.managers.control.run_supervisor_command') as run_mock:
        with patch('packy_agent.domain_logic.managers.control.Process') as process_cls_mock:
            with patch('packy_agent.domain_logic.managers.control.'
                       'control_manager.terminate_process'):
                with patch('packy_agent.domain_logic.managers.control.'
                           'control_manager.kill_process'):
                    with override_settings({'process_stop_timeout_seconds': 0.2,
                                            'sigterm_timeout_seconds': 0.2,
                                            'process_status_check_period_seconds': 0.02}):

                        terminate_at = time.time() + 0.5

                        def process_is_running_side_effect():
                            if time.time() >= terminate_at:
                                return False

                            return True

                        process_mock = MagicMock()
                        process_mock.is_running.side_effect = process_is_running_side_effect
                        process_cls_mock.return_value = process_mock

                        def side_effect(command, cargs, *args, **kwargs):
                            if command == 'status':
                                if time.time() >= terminate_at:
                                    return f'{cargs[0]} STOPPED'
                                else:
                                    return f'{cargs[0]} RUNNING pid 11, uptime 1 minute'

                        run_mock.side_effect = side_effect
                        control_manager.stop('worker')
                        run_mock.assert_has_calls(
                            (call('stop', ('worker',), is_async=False,
                                  async_delay_seconds=None, async_dev_null=False),), any_order=True)
                        control_manager.terminate_process.assert_called()
                        control_manager.kill_process.assert_called()



import time
import logging
from random import random

from packy_agent.worker.loops.base.misc import CallableLoop, describe_callable
from packy_agent.configuration.settings import settings


logger = logging.getLogger(__name__)


class PeriodicLoop(CallableLoop):

    loop_type = 'periodic'

    def __init__(self, period, callable_=None, args=(), kwargs=None,
                 is_logged_iteration=True, jitter_factor=None, **mkwargs):
        super().__init__(callable_=callable_, args=args, kwargs=kwargs,
                         is_safe_iteration=False,
                         is_logged_iteration=is_logged_iteration, **mkwargs)

        self.period = period
        if jitter_factor is None:
            self.jitter_factor = settings.get_worker_periodic_task_jitter_factor()
        else:
            self.jitter_factor = jitter_factor

    def iteration(self):
        self.call()

    def iteration_wrapper(self):
        description = self.description
        start_time = time.time()

        jitter_factor = self.jitter_factor
        period = self.period * (1 - jitter_factor + 2 * jitter_factor * random())

        try:
            super().iteration_wrapper()
        except Exception:
            elapsed = time.time() - start_time
            logger.exception(f'EXCEPTION in {description} job in {elapsed:.6g}s')
        else:
            elapsed = time.time() - start_time
            logger.info(f'DONE {description} job in {elapsed:.6g}s')

        if elapsed < period:
            wait_time = period - elapsed
            logger.debug(f'WAITING {wait_time:.6g}s for next iteration of {description}')
            self.sleep(wait_time)
        elif elapsed > period:
            longer = elapsed - period
            logger.warning(f'NOT WAITING {description} for next iteration: '
                           f'job took {longer:.6g}s longer than period ({period:.6g} seconds)')

    @property
    def description(self):
        if self.callable:
            part1 = describe_callable(self.callable, self.args, self.kwargs)
        else:
            part1 = self.__class__.__name__

        return f'{part1} (every {self.period}s)'



import json
import time

from schematics import Model, types


class CustomSchematicsModel(Model):

    def to_json(self, *args, **kwargs):
        json_kwargs = kwargs.pop('json_kwargs', {})
        return json.dumps(self.to_primitive(*args, **kwargs), **json_kwargs)

    def __repr__(self):
        return f'{self.__class__.__name__}({self.to_primitive()!r})'


class BaseMeasurement(CustomSchematicsModel):
    ts = types.IntType(default=lambda: int(time.time()), required=True)



import os

import pytest

from packy_agent.configuration.sources.key_value_storage import KeyValueStorageBackedSource

DATABASE_FILENAME = '/tmp/packy-agent-test-settings-database.sqlite3'
TABLENAME = 'test_kv'


@pytest.fixture
def key_value_storage_source():

    try:
        os.remove(DATABASE_FILENAME)
    except FileNotFoundError:
        pass

    return KeyValueStorageBackedSource(TABLENAME, DATABASE_FILENAME)



import logging
from urllib.parse import urljoin

import requests

from packy_agent.configuration.settings import settings
from packy_agent.utils.container import container


logger = logging.getLogger(__name__)


class PackyAgentWorkerClient:

    def get_base_url(self):
        return 'http://{}:{}'.format(settings.get_worker_http_bind_address(),
                                     settings.get_worker_http_port())

    def get_status(self):
        url = urljoin(self.get_base_url(), '/status/')
        logger.debug('Getting Worker status at %s', url)

        response = requests.get(url)
        response.raise_for_status()
        return response.json()


def get_packy_agent_worker_client():
    client = getattr(container, 'packy_agent_worker_client', None)
    if not client:
        container.packy_agent_worker_client = client = PackyAgentWorkerClient()

    return client



import gevent


from packy_agent.worker.loops.main import MainLoop


def start_main_loop(tasks):
    main_loop = MainLoop(tasks)
    gevent.spawn(main_loop.run)
    gevent.sleep(1)
    assert main_loop.is_running
    return main_loop


def stop_main_loop(main_loop, timeout=2):
    main_loop.stop()
    gevent.sleep(timeout)
    assert not main_loop.is_running


def run_main_loop(tasks, stop_tasks=True, duration_seconds=2, stop_timeout=2):
    main_loop = start_main_loop(tasks)
    gevent.sleep(duration_seconds)
    if stop_tasks:
        for task in main_loop.tasks:
            task.stop()

        gevent.sleep(stop_timeout)
        for task in main_loop.tasks:
            assert not task.is_running

    stop_main_loop(main_loop, stop_timeout)



import time
import logging

from packy_agent.worker.loops.main import MainLoop
from packy_agent.configuration.settings import settings
from packy_agent.clients.packy_server import get_packy_server_client


logger = logging.getLogger(__name__)


class Worker:

    def __init__(self):
        self.main_loop = None

    def run(self):
        logger.info('STARTED worker')
        if not settings.is_activated():
            logger.info('WAITING for agent activation...')

            activation_notify_period_seconds = settings.get_worker_activation_notify_period_seconds()
            while not settings.is_activated():
                get_packy_server_client().notify_server(settings.get_console_base_url())
                time.sleep(activation_notify_period_seconds)

        if settings.is_worker_stopped():
            logger.info('WAITING worker to be started...')

            started_check_period_seconds = settings.get_worker_started_check_period_seconds()
            while settings.is_worker_stopped():
                time.sleep(started_check_period_seconds)

        self.main_loop = MainLoop()
        self.main_loop.run()

        logger.info('STOPPED worker')

    def stop(self):
        logger.info('STOP requested for worker')
        if self.main_loop:
            self.main_loop.stop()



import logging

from packy_agent.configuration.settings import settings
from packy_agent.worker.loops.base.periodic import PeriodicLoop
from packy_agent.domain_logic.managers.control import control_manager
from packy_agent.domain_logic.managers.install_and_upgrade import refresh_access_token
from packy_agent.constants import GUARD_LOOP

logger = logging.getLogger(__name__)


def guard():
    if settings.get_access_token() in (settings.get_invalid_access_tokens() or ()):
        refresh_access_token()

    if settings.has_settings_changed_on_server():
        logger.warning('Apparently we missed reload event, will handle it now')
        # Although we apply new settings on fly we still need to restart everything in case
        # settings that are only once on start are changed
        control_manager.restart_all()


class GuardLoop(PeriodicLoop):

    formal_name = GUARD_LOOP

    def __init__(self, period):
        super().__init__(period=period, callable_=guard)



import logging

from sentry_sdk import capture_message

from packy_agent.constants import TRACE_MODULE, TRACE_MODULE_LOOP
from packy_agent.domain_logic.modules.traceroute.base import traceroute as traceroute_impl
from packy_agent.domain_logic.modules.traceroute.constants import UDP_METHOD
from packy_agent.worker.loops.modules.base import (
    ScheduledProducerTaskLoop, get_target, get_packet_size, get_number_of_pings)
from packy_agent.domain_logic.models.schematics.traceroute import TraceModuleMeasurement, Hop

HOP_ATTRIBUTES = ('last', 'best', 'worst', 'average', 'stdev')

logger = logging.getLogger(__name__)


def convert_to_hops(traceroute_result, probe_count):
    hops = []
    for x, item in enumerate(traceroute_result):
        hop_primitive = {
            'number': x + 1,

            'reply_number': item.get('reply_hop_number'),
            'ip_address': item.get('host'),

            'loss_abs': item['loss'],
            'sent': probe_count,
        }
        for attr in HOP_ATTRIBUTES:
            value = item.get(attr)
            if value is not None:
                hop_primitive[attr] = round(value * 1000.0, 1)

        hops.append(Hop(hop_primitive))

    return hops


def traceroute(*args, **kwargs):
    capture_message(f'Traceroute started with {args!r}, {kwargs!r}')

    target = get_target(kwargs, TRACE_MODULE)
    if not target:
        return

    packet_size_bytes = get_packet_size(kwargs, TRACE_MODULE)
    probe_count = get_number_of_pings(kwargs, TRACE_MODULE, default=3)

    measurement = TraceModuleMeasurement({
        'target': target,
        'packet_size': packet_size_bytes,
        'pings': probe_count,
        'hops': [],
    })  # we place it here because it also saves time moment

    if not (28 <= packet_size_bytes <= 1500):
        packet_size_bytes = min(max(28, packet_size_bytes), 1500)
        logger.warning('`packet_size` is out of range from 28 to 1500 (inclusive), '
                       'value is adjusted to {}'.format(packet_size_bytes))

    result = traceroute_impl(target,
                             probe_count=probe_count,
                             packet_size_bytes=packet_size_bytes,
                             max_hops=kwargs.get('ttl', 30),
                             method=kwargs.get('method', UDP_METHOD),
                             max_parallelism=kwargs.get('parallelism', 16))
    if result is None:
        logger.debug('traceroute_impl(%s) returned None', target)
        return

    capture_message(f'Got traceroute result: {result!r}')
    measurement.hops = convert_to_hops(result, probe_count)
    capture_message(f'Converted hops to: {measurement.hops!r}')
    return measurement


class TracerouteTaskLoop(ScheduledProducerTaskLoop):

    formal_name = TRACE_MODULE_LOOP

    def __init__(self, schedule, args=(), kwargs=None, outbound_queue=None):
        super().__init__(schedule=schedule, callable_=traceroute, args=args, kwargs=kwargs,
                         outbound_queue=outbound_queue)



import time

from unittest.mock import patch, call, MagicMock

from packy_agent.tests.fixtures.configuration import override_settings
from packy_agent.domain_logic.managers.control import control_manager


def test_restart():
    with patch('packy_agent.domain_logic.managers.control.run_supervisor_command') as run_mock:
        with patch('packy_agent.domain_logic.managers.control.Process') as process_cls_mock:
            with patch('packy_agent.domain_logic.managers.control.control_manager.stop'):
                process_mock = MagicMock()
                process_mock.is_running.return_value = False
                process_cls_mock.return_value = process_mock

                def side_effect(command, cargs, *args, **kwargs):
                    return (f'{cargs[0]} RUNNING pid 11, uptime 1 minute'
                            if command == 'status' else None)

                run_mock.side_effect = side_effect
                control_manager.restart('worker')
                run_mock.assert_has_calls(
                    (call('restart', ('worker',), is_async=False,
                          async_delay_seconds=None, async_dev_null=False),), any_order=True)
                control_manager.stop.assert_not_called()


def test_non_graceful_restart():
    with patch('packy_agent.domain_logic.managers.control.run_supervisor_command') as run_mock:
        with patch('packy_agent.domain_logic.managers.control.Process') as process_cls_mock:
            with patch('packy_agent.domain_logic.managers.control.'
                       'control_manager.terminate_process'):
                with patch('packy_agent.domain_logic.managers.control.'
                           'control_manager.kill_process'):
                    with patch('packy_agent.domain_logic.managers.control.control_manager.start'):
                        with override_settings({'process_restart_timeout_seconds': 0.2,
                                                'process_status_check_period_seconds': 0.02}):

                            terminate_at = time.time() + 0.5

                            def process_is_running_side_effect():
                                if time.time() >= terminate_at:
                                    return False

                                return True

                            process_mock = MagicMock()
                            process_mock.is_running.side_effect = process_is_running_side_effect
                            process_cls_mock.return_value = process_mock

                            def side_effect(command, cargs, *args, **kwargs):
                                if command == 'status':
                                    if time.time() >= terminate_at:
                                        return f'{cargs[0]} STOPPED'
                                    else:
                                        return f'{cargs[0]} RUNNING pid 11, uptime 1 minute'

                            run_mock.side_effect = side_effect
                            control_manager.restart('worker')
                            run_mock.assert_has_calls(
                                (call('restart', ('worker',), is_async=False,
                                      async_delay_seconds=None, async_dev_null=False),
                                 call('stop', ('worker',), is_async=False,
                                      async_delay_seconds=None, async_dev_null=False),),
                                any_order=True)
                            control_manager.start.assert_called()



import logging

from packy_agent.console.app import get_app
from packy_agent.utils.cli import get_base_argument_parser
from packy_agent.utils.logging import configure_logging_basic
from packy_agent.configuration.settings import settings
from packy_agent.domain_logic.constants import CONSOLE_COMPONENT


def entry():
    settings.set_runtime('component', CONSOLE_COMPONENT)

    configure_logging_basic(logging.WARNING)
    parser = get_base_argument_parser(__loader__.name)

    parser.add_argument('--host', default='127.0.0.1')
    parser.add_argument('--http-port', '--port', type=int, default=settings.get_console_http_port())
    parser.add_argument('--debug', action='store_true')

    args = parser.parse_args()

    app = get_app(args)
    app.run(host=args.host, port=settings.get_console_http_port(), debug=args.debug)


if __name__ == '__main__':
    entry()



from packy_agent.worker.loops.base.consumer import ConsumerLoop
from packy_agent.constants import CONSUMER_LOOP
from packy_agent.clients.local_storage import get_local_storage


class TaskResultsConsumer(ConsumerLoop):

    formal_name = CONSUMER_LOOP

    def __init__(self, inbound_queue=None, timeout=None):
        super().__init__(inbound_queue, timeout=timeout)

        self.collected_counter = 0

    def process_item(self, item):
        get_local_storage().save_measurement(item)
        self.collected_counter += 1



import logging
import os.path

from packy_agent.worker.loops.base.periodic import PeriodicLoop
from packy_agent.clients.local_storage import get_local_storage
from packy_agent.configuration.settings import settings
from packy_agent.utils.fs import get_available_disk_space
from packy_agent.constants import PURGER_LOOP

MIN_SPACE_TO_FREE = 8192


logger = logging.getLogger(__name__)


def purge():
    local_storage = get_local_storage()
    deleted_counter = local_storage.purge_submitted_measurements()
    deleted_counter += local_storage.purge_failed_measurements()

    logger.debug('Evaluating a need to purge unsubmitted measurements...')

    minimum_disk_space_bytes = settings.get_minimum_disk_space_bytes()
    available_disk_space = get_available_disk_space()
    if available_disk_space < minimum_disk_space_bytes:
        logger.debug('Need to purge unsubmitted measurements')
        to_free = minimum_disk_space_bytes - available_disk_space
        if to_free >= MIN_SPACE_TO_FREE:
            number_of_records = local_storage.get_measurements_number()
            logger.debug('Found %s unsubmitted measurements', number_of_records)
            if number_of_records:
                file_size = os.path.getsize(settings.get_database_filename())
                bytes_per_record = file_size / number_of_records
                records_to_delete = int(to_free / bytes_per_record)
                logger.debug('Will purge %s unsubmitted measurements', records_to_delete)
                deleted_counter += local_storage.purge_unsubmitted_measurements(
                    records_to_delete)

    return deleted_counter


class TaskResultsPurger(PeriodicLoop):

    formal_name = PURGER_LOOP

    def __init__(self):
        super().__init__(settings.get_worker_purge_period_seconds())
        self.purged_records = 0

    def call(self):
        purged_records = purge()
        self.purged_records += purged_records




import logging
import socket

from wampy.message_handler import MessageHandler
from wampy.transports.websocket.connection import WebSocket, SecureWebSocket


logger = logging.getLogger(__name__)


class TicketAuthenticate:
    WAMP_CODE = 5
    name = 'authenticate'

    def __init__(self, ticket, kwargs_dict=None):
        super(TicketAuthenticate, self).__init__()

        self.ticket = ticket
        self.kwargs_dict = kwargs_dict or {}

    @property
    def message(self):
        return [self.WAMP_CODE, self.ticket, self.kwargs_dict]


class TicketMessageHandler(MessageHandler):

    ticket = None

    def __init__(self, ticket) -> None:
        super().__init__()
        self.ticket = ticket

    def handle_challenge(self, message_obj):
        logger.debug('client has been Challenged')
        message = TicketAuthenticate(self.ticket)
        self.session.send_message(message)


class CustomWebSocket(WebSocket):
    def disconnect(self):
        if self.socket:
            try:
                self.socket.shutdown(socket.SHUT_RDWR)
            except socket.error:
                pass

            self.socket.close()


class CustomSecureWebSocket(SecureWebSocket, CustomWebSocket):
    pass



import logging
import base64
import socket
from functools import wraps
from urllib.parse import urljoin, urlencode

import requests
from requests.exceptions import ConnectionError as RequestsConnectionError

import packy_agent
from packy_agent.exceptions import ImproperlyConfiguredError, NoAccessTokenError
from packy_agent.configuration.settings import settings
from packy_agent.utils.misc import get_lazy_value
from packy_agent.utils.platforms import get_platform
from packy_agent.exceptions import AuthenticationError
from packy_agent.utils.collections import set_if
from packy_agent.utils.container import container
from packy_agent.utils.data_usage import get_network_usage


logger = logging.getLogger(__name__)

OAUTH_TOKEN_PATH = 'oauth/token/'

API_V2_BASE_URL = 'api/v2/'

PENDING_AGENT_PATH = API_V2_BASE_URL + 'pending-agent/'
AGENT_VERSION_PATH = API_V2_BASE_URL + 'agent-version/'
USER_AGENT_TEMPLATE_PATH = API_V2_BASE_URL + 'user/agent/'
AGENT_TEMPLATE_PATH = API_V2_BASE_URL + 'agent/'
MEASUREMENT_TEMPLATE_PATH = AGENT_TEMPLATE_PATH + 'module/{}/measurement/'
AGENT_INSTANCE_TEMPLATE_PATH = USER_AGENT_TEMPLATE_PATH + '{}/'


METHOD_MAP = {
    'GET': requests.get,
    'POST': requests.post,
    'PATCH': requests.patch,
}

GET_INACTIVE_AGENTS_PARAMS = urlencode({'fields': 'id,name', 'is_active': 'false'})


def make_basic_authorization_value(username, password):
    return 'Basic ' + base64.b64encode((username + ':' + password).encode()).decode()


def make_token_authorization_value(access_token):
    return 'Bearer ' + access_token


def generate_agent_name():
    return socket.gethostname()


def if_callable(value):
    return value is not None


def with_agent_access_token(func_or_meth):
    @wraps(func_or_meth)
    def wrapper(*args, **kwargs):
        if not settings.get_access_token():
            raise NoAccessTokenError()
        return func_or_meth(*args, **kwargs)

    return wrapper


class PackyServerClient:
    # TODO(dmu) HIGH: Implement support for persistent HTTP?

    def __init__(self, base_url=None, timeout=None):
        self._base_url = base_url or settings.get_server_base_url
        self._timeout = timeout or settings.get_http_timeout_seconds

        self._user_agent = 'packy-agent/{} ({})'.format(packy_agent.__version__, get_platform())

    @property
    def base_url(self):
        base_url = get_lazy_value(self._base_url)
        if not base_url:
            raise ImproperlyConfiguredError('Packy Server base URL is not configured')
        return base_url

    @property
    def timeout(self):
        return get_lazy_value(self._timeout)

    def make_url(self, path):
        return urljoin(self.base_url, path)

    # TODO(dmu) MEDIUM: Get rid of `auto_login`. Support multiple authentication methods instead
    def request(self, url, data=None, json=None, method=None, headers=None,
                authorization=None, raise_for_status=True):

        if not method:
            if data is not None or json is not None:
                method = 'POST'
            else:
                method = 'GET'

        python_method = METHOD_MAP.get(method)
        if not python_method:
            raise NotImplementedError('Method {} is not supported'.format(method))

        if not authorization:
            access_token = settings.get_access_token()
            if access_token:
                authorization = make_token_authorization_value(access_token)

        # TODO(dmu) HIGH: Implement token expiration support
        # if auto_login and (not self.token or time.time() >= self.token_expiration):
        #     self.login()

        headers = headers.copy() if headers else {}
        if authorization:
            headers.setdefault('Authorization', authorization)
        headers.setdefault('Content-Type', 'application/json')
        headers.setdefault('User-Agent', self._user_agent)

        logger.debug(f'HTTP {method} {url} "{data or json}" request')
        response = python_method(url, data=data, json=json, headers=headers, timeout=self.timeout)

        if response.status_code >= 400:
            logger.debug('HTTP%s: %s', response.status_code, response.content)

        if raise_for_status:
            response.raise_for_status()

        return response

    # E-mail/password authentication methods
    ########################################

    def get_inactive_agents(self, email, password, raise_for_status=True):
        return self.request(
            self.make_url(USER_AGENT_TEMPLATE_PATH) + '?' + GET_INACTIVE_AGENTS_PARAMS,
            authorization=make_basic_authorization_value(email, password),
            raise_for_status=raise_for_status)

    def create_agent(self, email, password, raise_for_status=True):
        payload = {'name': generate_agent_name()}
        return self.request(self.make_url(USER_AGENT_TEMPLATE_PATH), json=payload,
                            authorization=make_basic_authorization_value(email, password),
                            raise_for_status=raise_for_status)

    def activate_agent(self, agent_id, authorization, raise_for_status=True):
        url = (self.make_url(AGENT_INSTANCE_TEMPLATE_PATH).format(agent_id) + '?' +
               urlencode({'with_configuration': 'yes'}))
        return self.request(url, json={'is_active': True}, method='PATCH',
                            authorization=authorization, raise_for_status=raise_for_status)

    def activate_agent_basic_auth(self, email, password, agent_id, raise_for_status=True):
        return self.activate_agent(agent_id=agent_id,
                                   authorization=make_basic_authorization_value(email, password),
                                   raise_for_status=raise_for_status)

    def validate_auth(self, email, password, agent_key=None):
        agent_key = agent_key or settings.get_agent_key()
        if not agent_key:
            raise ImproperlyConfiguredError('`agent_key` must be provided')

        # Use filter to get the agent by key
        url = self.make_url(
            USER_AGENT_TEMPLATE_PATH) + '?' + urlencode(
                {'fields': 'key,user', 'expand': 'user', 'key': agent_key})
        response = self.request(url, authorization=make_basic_authorization_value(email, password),
                                raise_for_status=False)
        status_code = response.status_code

        # Credentials may be invalid themselves (i.e. typo in password)
        if status_code == 401:
            logger.info('Invalid credentials')
            return False

        # There may be other errors
        response.raise_for_status()

        # Server should respond with a list, but in case it does not we take care
        # (this should never happen).
        response_json = response.json()
        if not isinstance(response_json, list):
            logger.warning('Server responded with {!r} instead of a list'.format(response_json))
            return False

        # List is empty. This may mean that agent was deleted on server or that credentials
        # provided belong to another user (not the one agent belongs to)
        if not response_json:
            return False

        # List is longer than 1 item. Something went wrong on server (this should never happen).
        if len(response_json) > 1:
            logger.warning('Server returned multiple agents instead of one')
            return False

        agent_json = response_json[0]
        # Double check that we got email we expected
        if (agent_json.get('user') or {}).get('email') != email:
            logger.warning('Server returned agent of another user')
            return False

        # Double check that we got key we expected
        if agent_json['key'] != agent_key:
            logger.warning('Server returned another agent')
            return False

        return True

    # Client secret authentication methods
    ######################################
    def get_access_token_for_credentials(self, client_id, client_secret):
        # TODO(dmu) HIGH: Remove this method once all agents migrate to v2
        url = self.make_url(OAUTH_TOKEN_PATH)
        response = self.request(url, data={'grant_type': 'client_credentials',
                                           'client_id': client_id,
                                           'client_secret': client_secret},
                                headers={'Content-Type': 'application/x-www-form-urlencoded'},
                                raise_for_status=False)
        if response.status_code == 200:
            response_json = response.json()
            return response_json['access_token']
        else:
            raise AuthenticationError('Not authenticated: %s', response.content)

    # Refresh token authentication methods
    ######################################
    def refresh_access_token(self, client_id, refresh_token):
        url = self.make_url(OAUTH_TOKEN_PATH)
        response = self.request(url, data={'grant_type': 'refresh_token',
                                           'client_id': client_id,
                                           'refresh_token': refresh_token},
                                headers={'Content-Type': 'application/x-www-form-urlencoded'},
                                raise_for_status=False)
        status_code = response.status_code
        if status_code == 200:
            return response.json()
        elif status_code == 401:
            raise AuthenticationError('Not authenticated: %s', response.content)
        else:
            response.raise_for_status()

    # No authentication methods
    ###########################
    def notify_server(self, agent_console_url):
        url = self.make_url(PENDING_AGENT_PATH)
        payload = {
            'agent_control_server_url': agent_console_url,  # TODO(dmu) LOW: Legacy, remove
            'agent_console_url': agent_console_url,
        }
        try:
            requests.post(url, json=payload, timeout=self.timeout)
        except RequestsConnectionError:
            logger.info('Unable to notify: Packy Server unavailable')

    def get_version_max(self):
        url = self.make_url(AGENT_VERSION_PATH)
        response = self.request(url)
        return response.json().get('version_max')

    # Access token authentication methods
    #####################################
    def activate_agent_token_auth(self, access_token, agent_id, raise_for_status=True):
        return self.activate_agent(agent_id=agent_id,
                                   authorization=make_token_authorization_value(access_token),
                                   raise_for_status=raise_for_status)

    def deactivate_agent_legacy(self, access_token, agent_id):
        return self.request(self.make_url(AGENT_INSTANCE_TEMPLATE_PATH.format(agent_id)),
                            json={'is_active': False}, method='PATCH',
                            authorization=make_token_authorization_value(access_token))

    @with_agent_access_token
    def deactivate_agent(self):
        return self.request(self.make_url(AGENT_TEMPLATE_PATH),
                            json={'is_active': False}, method='PATCH')

    @with_agent_access_token
    def submit_measurement(self, module_public_identifier, measurement, raise_for_status=True):
        url = self.make_url(MEASUREMENT_TEMPLATE_PATH.format(module_public_identifier))
        return self.request(url, data=measurement, raise_for_status=raise_for_status)

    @with_agent_access_token
    def update_status(self, version, ip_address=None, is_online=True):
        payload = {
            'is_online': is_online,
            'version': version,
        }

        set_if(payload, 'ip_address', ip_address)

        bytes_sent, bytes_received = get_network_usage()
        set_if(payload, 'bytes_sent', bytes_sent, if_callable=if_callable)
        set_if(payload, 'prev_bytes_sent', settings.get('bytes_sent'), if_callable=if_callable)
        set_if(payload, 'bytes_received', bytes_received, if_callable=if_callable)
        set_if(payload, 'prev_bytes_received', settings.get('bytes_received'),
               if_callable=if_callable)

        url = self.make_url(AGENT_TEMPLATE_PATH)
        try:
            response = self.request(url, json=payload, method='PATCH', raise_for_status=False)
        except RequestsConnectionError:
            logger.info('Unable to update status (send heartbeat): Packy Server unavailable')
            return
        else:
            status_code = response.status_code
            if status_code == 200:
                settings.set('bytes_sent', bytes_sent)
                settings.set('bytes_received', bytes_received)
            elif status_code == 401:
                logger.debug('Access token is invalid')
                # We do not expect context switch here because we work with RAM only (no IO)
                invalid_access_tokens = settings.get_invalid_access_tokens() or set()
                invalid_access_tokens.add(settings.get_access_token())
                settings.set_runtime('invalid_access_tokens', invalid_access_tokens)
            elif status_code >= 500:
                logger.info('Packy Server responded with HTTP%s', status_code)
            elif status_code >= 400:
                response.raise_for_status()
            else:
                logger.warning('Unexpected status while sending heartbeat: HTTP%s', status_code)

        return response

    @with_agent_access_token
    def get_settings(self):
        url = self.make_url(AGENT_TEMPLATE_PATH) + '?fields=settings'
        response = self.request(url)
        return response.json()['settings']

    @with_agent_access_token
    def is_agent_online(self):
        url = self.make_url(AGENT_TEMPLATE_PATH) + '?fields=is_online'
        response = self.request(url)
        response.raise_for_status()
        return response.json().get('is_online')


def get_inactive_agents(email, password):
    response = get_packy_server_client().get_inactive_agents(email, password, raise_for_status=False)

    if response.status_code == 401:
        raise AuthenticationError('Invalid credentials for agent activation')

    response.raise_for_status()
    return {agent['id']: agent['name'] for agent in response.json()}


def get_packy_server_client():
    client = getattr(container, 'packy_server_client', None)
    if not client:
        container.packy_server_client = client = PackyServerClient()

    return client



import pytest

from urllib.parse import urlparse


def test_index_page_smoke(console_flask_client):
    response = console_flask_client.get('/')
    assert response.status_code == 302
    assert urlparse(response.location).path == '/activate/'


@pytest.mark.usefixtures('activated')
def test_index_page_activated_agent(console_flask_client):
    response = console_flask_client.get('/')
    assert response.status_code == 302
    assert urlparse(response.location).path == '/login/'



import time
from unittest.mock import MagicMock

import pytest
from gevent.queue import Queue

from packy_agent.worker.loops.modules.ping import ping, PingTaskLoop
from packy_agent.domain_logic.models.schematics.ping import PingModuleMeasurement
from packy_agent.exceptions import ImproperlyConfiguredError


def test_ping_smoke():
    measurement = ping(target='8.8.8.8')
    assert isinstance(measurement, PingModuleMeasurement)


def test_ping_smoke_host():
    measurement = ping(host='8.8.8.8')
    assert isinstance(measurement, PingModuleMeasurement)


def test_ping():
    measurement = ping(target='8.8.8.8')
    assert isinstance(measurement, PingModuleMeasurement)
    assert measurement.target == '8.8.8.8'
    assert measurement.packet_size == 64
    assert measurement.n_pings == 10
    assert isinstance(measurement.values_, list)
    assert measurement.values_
    for value in measurement.values_:
        assert isinstance(value, float)


def test_ping_domain_name():
    measurement = ping(target='google.com')
    assert isinstance(measurement, PingModuleMeasurement)
    assert measurement.target == 'google.com'
    assert measurement.packet_size == 64
    assert measurement.n_pings == 10
    assert isinstance(measurement.values_, list)
    assert measurement.values_
    for value in measurement.values_:
        assert isinstance(value, float)


def test_ping_unresolvable_target():
    measurement = ping(target='blahblahblah')
    assert isinstance(measurement, PingModuleMeasurement)
    assert isinstance(measurement.values_, list)
    assert not measurement.values_


def test_ping_with_interval():
    measurement = ping(target='8.8.8.8', interval_ms=50)
    assert isinstance(measurement, PingModuleMeasurement)
    assert measurement.target == '8.8.8.8'
    assert measurement.packet_size == 64
    assert measurement.n_pings == 10
    assert isinstance(measurement.values_, list)
    assert measurement.values_
    for value in measurement.values_:
        assert isinstance(value, float)


def test_ping_no_target_provided():
    assert ping() is None


def test_ping_task_loop_puts_to_queue():
    queue = Queue()
    loop = PingTaskLoop('*/1 * * * * *', kwargs={'target': '8.8.8.8'}, outbound_queue=queue)
    loop.croniter = MagicMock()
    loop.croniter.get_next.return_value = time.time() + 0.1
    loop.iteration_wrapper()

    assert len(queue) == 1
    measurement = queue.get()
    assert isinstance(measurement, PingModuleMeasurement)
    assert measurement.target == '8.8.8.8'
    assert measurement.packet_size == 64
    assert measurement.n_pings == 10
    assert isinstance(measurement.values_, list)
    assert measurement.values_
    for value in measurement.values_:
        assert isinstance(value, float)



import gevent
from packy_agent.tests.worker.loops.main.base import start_main_loop, stop_main_loop


TASKS = [
    {
        'module_name': 'ping', 'args': [], 'id': 1,
        'kwargs': {
            'number_of_pings': 3,
            'override': False,
            'host': '8.8.8.8',
            'interval_ms': 0,
            'packet_size': 64
        },
        'schedule': '*/1 * * * * *'
    },
]


def test_main_loop_graceful_stop():
    main_loop = start_main_loop(TASKS)
    stop_main_loop(main_loop)



import json

import pytest
import httpretty

from urllib.parse import urlparse
from packy_agent.configuration.settings import settings


def test_activation(console_flask_client):
    assert not settings.is_activated()

    response = console_flask_client.get('/activate/')
    assert response.status_code == 200

    with httpretty.enabled():
        url = settings.get_server_base_url() + 'api/v2/user/agent/'
        httpretty.register_uri(httpretty.GET, url, body='[]')
        httpretty.register_uri(
            httpretty.POST, url, status=201,
            body=json.dumps({
                'access_token': 'access_token',
                'refresh_token': 'refresh_token',
                'key': 'key',
                'name': 'name',
                'client_id': 'client_id',
            }))

        response = console_flask_client.post(
            '/activate/', data={'email': 'test@test.com', 'password': 'test'},
            follow_redirects=True)
        assert response.status_code == 200

        latest_requests = httpretty.HTTPretty.latest_requests
        assert len(latest_requests) >= 2
        assert latest_requests[0].path == '/api/v2/user/agent/?fields=id%2Cname&is_active=false'
        assert latest_requests[0].method == 'GET'
        assert latest_requests[1].path == '/api/v2/user/agent/'
        assert latest_requests[1].method == 'POST'
        body = json.loads(latest_requests[1].body)
        assert 'name' in body
        assert isinstance(body['name'], str)

    assert settings.is_activated()


def test_activation_invalid_password(console_flask_client):
    assert not settings.is_activated()

    response = console_flask_client.get('/activate/')
    assert response.status_code == 200

    with httpretty.enabled():
        url = settings.get_server_base_url() + 'api/v2/user/agent/'
        httpretty.register_uri(httpretty.GET, url, status=401)

        response = console_flask_client.post(
            '/activate/', data={'email': 'test@test.com', 'password': 'test'})
        assert response.status_code == 302
        assert urlparse(response.location).path == '/activation-failure/'

    assert not settings.is_activated()


@pytest.mark.parametrize('http_status', (500, 501, 502, 503, 504))
def test_activation_server_http5xx(console_flask_client, http_status):
    assert not settings.is_activated()

    response = console_flask_client.get('/activate/')
    assert response.status_code == 200

    with httpretty.enabled():
        url = settings.get_server_base_url() + 'api/v2/user/agent/'
        httpretty.register_uri(httpretty.GET, url, status=http_status)

        response = console_flask_client.post(
            '/activate/', data={'email': 'test@test.com', 'password': 'test'})
        assert response.status_code == 302
        assert urlparse(response.location).path == '/activation-failure/'

    assert not settings.is_activated()


@pytest.mark.parametrize('http_status', (400, 401, 403, 404, 405, 406, 408, 429))
def test_activation_server_http4xx(console_flask_client, http_status):
    assert not settings.is_activated()

    response = console_flask_client.get('/activate/')
    assert response.status_code == 200

    with httpretty.enabled():
        url = settings.get_server_base_url() + 'api/v2/user/agent/'
        httpretty.register_uri(httpretty.GET, url, status=http_status)

        response = console_flask_client.post(
            '/activate/', data={'email': 'test@test.com', 'password': 'test'})
        assert response.status_code == 302
        assert urlparse(response.location).path == '/activation-failure/'

    assert not settings.is_activated()


def test_activation_server_connection_error(console_flask_client):
    assert not settings.is_activated()

    response = console_flask_client.get('/activate/')
    assert response.status_code == 200

    response = console_flask_client.post(
        '/activate/', data={'email': 'test@test.com', 'password': 'test'})
    assert response.status_code == 302
    assert urlparse(response.location).path == '/activation-failure/'

    assert not settings.is_activated()



import logging

import flask
from flask.views import MethodView
from packy_agent.utils.auth import logout


logger = logging.getLogger(__name__)


class LogoutView(MethodView):

    def post(self):
        logout()
        return flask.redirect(flask.url_for('login'))



import logging
from urllib.parse import urlparse
from datetime import datetime
import socket
import ssl


from packy_agent.domain_logic.models.schematics.http import HTTPModuleMeasurement


logger = logging.getLogger(__name__)


SSL_DT_FORMAT = r'%b %d %H:%M:%S %Y %Z'
CHROME_USERAGENT = ('Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) '
                    'Chrome/56.0.2924.87 Safari/537.36')


def get_certificate_expiration_dt(url):
    url_parsed = urlparse(url)
    if url_parsed.scheme != 'https':
        return None

    parts = (url_parsed.netloc + ':443').split(':')
    hostname = parts[0]
    port = int(parts[1])

    context = ssl.create_default_context()
    connection = context.wrap_socket(socket.socket(socket.AF_INET), server_hostname=hostname)
    connection.settimeout(5)

    connection.connect((hostname, port))
    ssl_info = connection.getpeercert()

    return datetime.strptime(ssl_info['notAfter'], SSL_DT_FORMAT)


def get_http_measurement(url, follow_redirects=False):
    logger.debug('Getting HTTP measurement for %s', url)

    # We import here to allow install/upgrade procedure successfully install pycurl system library
    # dependencies
    import pycurl
    curl_object = pycurl.Curl()

    curl_object.setopt(pycurl.URL, url)
    curl_object.setopt(pycurl.USERAGENT, CHROME_USERAGENT)
    if follow_redirects:
        curl_object.setopt(pycurl.MAXREDIRS, 50)
        curl_object.setopt(pycurl.FOLLOWLOCATION, 50)

    curl_object.setopt(pycurl.WRITEFUNCTION, lambda dummy: None)
    measurement = HTTPModuleMeasurement({'target': url})
    try:
        curl_object.perform()
    except pycurl.error:
        measurement.is_success = False
        return measurement

    measurement.http_status_code = curl_object.getinfo(pycurl.RESPONSE_CODE)
    measurement.namelookup_ms = round(curl_object.getinfo(pycurl.NAMELOOKUP_TIME) * 1000, 2)
    measurement.total_ms = round(curl_object.getinfo(pycurl.TOTAL_TIME) * 1000, 2)

    try:
        measurement.certificate_expiration_dt = get_certificate_expiration_dt(url)
    except IOError:
        logger.exception('Error while getting certification expiration datetime')

    return measurement



import json
from urllib.parse import urlparse

import flask
import httpretty
import pytest

from packy_agent.configuration.settings import settings
from packy_agent.tests.fixtures.configuration import override_settings


@pytest.mark.usefixtures('activated')
def test_login_logout(console_flask_client):
    agent_key = 'agent_key_value'
    assert settings.is_activated()

    response = console_flask_client.get('/')
    assert response.status_code == 302
    assert urlparse(response.location).path == '/login/'

    response = console_flask_client.get('/login/')
    assert response.status_code == 200

    with httpretty.enabled():
        url = settings.get_server_base_url() + 'api/v2/user/agent/'
        httpretty.register_uri(
            httpretty.GET, url,
            body=json.dumps([{'key': agent_key, 'user': {'email': 'test@test.com'}}]))

        response = console_flask_client.post(
            '/login/', data={'email': 'test@test.com', 'password': 'test'},
            follow_redirects=True)
        assert response.status_code == 200
        assert flask.session['agent_key'] == agent_key

        latest_requests = httpretty.HTTPretty.latest_requests
        assert len(latest_requests) == 1
        assert latest_requests[0].method == 'GET'
        assert (latest_requests[0].path ==
                f'/api/v2/user/agent/?fields=key%2Cuser&expand=user&key={agent_key}')

    response = console_flask_client.get('/')
    assert response.status_code == 200

    response = console_flask_client.post('/logout/', follow_redirects=True)
    assert response.status_code == 200
    assert 'agent_key' not in flask.session

    response = console_flask_client.get('/')
    assert response.status_code == 302
    assert urlparse(response.location).path == '/login/'



from setuptools import setup

version = __import__('packy_agent').__version__

setup(
    name='packy-agent',
    version=version,
    description='Packy Agent',
    packages=['packy_agent'],
    zip_safe=False,
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'packy-agent-console = packy_agent.console.run:entry',
            'packy-agent-watchdog = packy_agent.watchdog.run:entry',
            'packy-agent-worker = packy_agent.worker.run:entry',
            'packy-agent-settings = packy_agent.cli.settings:entry',
            'packy-agent-activate = packy_agent.cli.activate:entry',
            'packy-agent-welcome = packy_agent.cli.welcome:entry',
            'packy-agent-ping = packy_agent.cli.ping:entry',
            'packy-agent-traceroute = packy_agent.cli.traceroute:entry',
            'packy-agent-http = packy_agent.cli.http:entry',
            'packy-agent-set-dhcp = packy_agent.cli.set_dhcp:entry',
            'packy-agent-version = packy_agent.cli.version:entry',
            'packy-agent-migrate1to2 = packy_agent.cli.migrate1to2:entry',
        ]
    }
)



import logging
import os
import subprocess
import signal


logger = logging.getLogger(__name__)


def run_command_sync(command, raise_exception=False):
    logger.info('Started sync shell command: %s', command)
    try:
        output = subprocess.check_output(command, shell=True, stderr=subprocess.STDOUT).strip()
    except subprocess.CalledProcessError as ex:
        logger.error('Sync shell command failed: %s: %s', command, ex.output.strip())
        if raise_exception:
            raise
    else:
        logger.info('Sync shell command succeeded: %s: %s', command, output.strip())
        return output


def run_command_async(command, delay_seconds=None, dev_null=False):
    if delay_seconds is not None:
        command = f'sleep {delay_seconds}; {command}'

    command = f'({command}) &'

    if dev_null:
        command = f'({command}) > /dev/null 2>&1'

    logger.debug('Started async shell command: %s', command)
    os.system(command)
    logger.debug('Finished async shell command: %s', command)


def run_command(command, is_async=False, raise_exception=False, async_delay_seconds=None,
                async_dev_null=False):
    logger.info('RUNNING shell command: %s', command)
    if is_async:
        run_command_async(command, delay_seconds=async_delay_seconds, dev_null=async_dev_null)
    else:
        return run_command_sync(command, raise_exception=raise_exception)


def send_signal(pid, signal_no, is_async=False, async_delay_seconds=None,
                async_dev_null=False):
    if is_async:
        run_command(f'kill -{signal_no} {pid}', is_async=True,
                    async_delay_seconds=async_delay_seconds, async_dev_null=async_dev_null)
    else:
        os.kill(pid,  signal_no)


def reload_process(pid, is_async=False, async_delay_seconds=None, async_dev_null=False):
    send_signal(pid, signal.SIGHUP, is_async=is_async,
                async_delay_seconds=async_delay_seconds, async_dev_null=async_dev_null)


def terminate_process(pid, is_async=False, async_delay_seconds=None, async_dev_null=False):
    send_signal(pid, signal.SIGTERM, is_async=is_async,
                async_delay_seconds=async_delay_seconds, async_dev_null=async_dev_null)


def kill_process(pid, is_async=False, async_delay_seconds=None, async_dev_null=False):
    send_signal(pid, signal.SIGKILL, is_async=is_async,
                async_delay_seconds=async_delay_seconds, async_dev_null=async_dev_null)



import cachetools


def custom_cached(cache, key=cachetools.keys.hashkey, lock=None):
    def decorator(func):
        func.cache = cache
        return cachetools.cached(cache=cache, key=key, lock=lock)(func)

    return decorator



import binascii
import os
import logging
import random
import distutils
import cachetools

import yaml

from atomicwrites import AtomicWriter, atomic_write as atomic_write_original
from packy_agent.utils.cache import custom_cached


SENTINEL = object()


logger = logging.getLogger(__name__)


def yaml_coerce(value):
    if isinstance(value, str):
        return yaml.load('dummy: ' + value)['dummy']

    return value


@custom_cached(cachetools.TTLCache(100, ttl=1))
def get_lazy_value(value):
    return value() if hasattr(value, '__call__') else value


def get_executable_path(executable, self_failover=True):
    path = distutils.spawn.find_executable(executable)
    if not path and self_failover:
        path = executable

    return path


class CustomAtomicWriter(AtomicWriter):
    def commit(self, f):
        path = self._path
        if os.path.isfile(path):
            filestat = os.stat(path)
        else:
            filestat = None

        super(CustomAtomicWriter, self).commit(f)

        if filestat is not None:
            try:
                os.chmod(path, filestat.st_mode)
                os.chown(path, filestat.st_uid, filestat.st_gid)
            except Exception:
                logger.warning('Failed to restore permissions or ownership for file: %s', path)


def atomic_write(path, writer_cls=CustomAtomicWriter, **cls_kwargs):
    return atomic_write_original(path=path, writer_cls=writer_cls, **cls_kwargs)


def generate_random_string(size_bytes):
    return (('{:0' + str(size_bytes) + 'x}').format(random.getrandbits(size_bytes * 4))
            if size_bytes else '')


def generate_flask_secret_key():
    return generate_random_string(128)


def safe_hexlify(data):
    return binascii.hexlify(data).decode() if data else data



import time
import logging

from sqlalchemy import create_engine, delete, select, func

from packy_agent.configuration.settings import settings
from packy_agent.utils.container import container
from packy_agent.utils.gevent import database_locks
from packy_agent.domain_logic.models.sqlalchemy.base import Session
from packy_agent.domain_logic.models.sqlalchemy.measurement import Measurement
from packy_agent.domain_logic.models.schematics.ping import PingModuleMeasurement
from packy_agent.domain_logic.models.schematics.traceroute import TraceModuleMeasurement
from packy_agent.domain_logic.models.schematics.speedtest import SpeedtestModuleMeasurement
from packy_agent.domain_logic.models.schematics.http import HTTPModuleMeasurement
from packy_agent.constants import ErrorSide

TYPE_MAP = {
    PingModuleMeasurement: 1,
    TraceModuleMeasurement: 2,
    SpeedtestModuleMeasurement: 3,
    HTTPModuleMeasurement: 4,
}

REVERSE_TYPE_MAP = {v: k for k, v in TYPE_MAP.items()}
assert len(TYPE_MAP) == len(REVERSE_TYPE_MAP)

logger = logging.getLogger(__name__)


class LocalStorage:

    def __init__(self):
        self._engine = None

    @property
    def engine(self):
        if not self._engine:
            self._engine = create_engine(settings.get_database_url())

        return self._engine

    def save_measurement(self, instance):
        # TODO(dmu) LOW: Get database url from self.engine
        with database_locks[settings.get_database_url()]:
            with Session(self.engine) as session:
                session.add(
                    Measurement(
                        measurement_type=TYPE_MAP[instance.__class__],
                        value=instance.to_json(json_kwargs={'separators': (',', ':')}).encode()
                    ))

    def get_remaining_measurements(self, measurement_type=None, limit=100):
        with Session(self.engine) as session:
            query = session.query(Measurement).filter_by(submitted_at_ts=None)
            if measurement_type is not None:
                query = query.filter_by(measurement_type=measurement_type)
            return query.order_by(Measurement.created_at_ts)[:limit]

    def mark_measurement_as_submitted(self, id_):
        logger.debug(f'MARKING measurement as submitted: id = {id_!r}')
        with Session(self.engine) as session:
            session.query(Measurement).filter_by(id=id_).update(
                values={'submitted_at_ts': time.time()})
        logger.debug(f'MARKED measurement as submitted: id = {id_!r}')

    def save_measurement_error(self, id_, error_side, error_message=None, error_at_ts=None):
        logger.debug(f'SAVING measurement error: id = {id_!r}, side = {error_side}, '
                     f'error_message = {error_message!r}, error_at_ts = {error_at_ts!r} ...')

        values = {
            'error_side': error_side,
            'error_at_ts': error_at_ts or time.time(),
            'error_message': error_message
        }
        with Session(self.engine) as session:
            session.query(Measurement).filter_by(id=id_).update(values=values)

        logger.debug(f'SAVED measurement error: id = {id_!r}, side = {error_side}, '
                     f'error_message = {error_message!r}, error_at_ts = {error_at_ts!r}')

    def purge_measurements(self, filter_args=None, filter_by_kwargs=None):
        with Session(self.engine) as session:
            qs = session.query(Measurement)
            if filter_args:
                qs = qs.filter(*filter_args)

            if filter_by_kwargs:
                qs = qs.filter_by(**filter_by_kwargs)

            deleted_count = qs.delete()

        if deleted_count:
            logger.debug('STARTED VACUUM')
            self.engine.execute('VACUUM')
            logger.debug('FINISHED VACUUM')

        return deleted_count

    def purge_unsubmitted_measurements(self, number_of_records=None):
        if number_of_records:
            inner_query = select(
                [Measurement.id]).order_by(Measurement.created_at_ts).limit(number_of_records)
            query = delete(Measurement).where(Measurement.id.in_(inner_query))
        else:
            query = delete(Measurement)

        with Session(self.engine) as session:
            return session.execute(query).rowcount

    def purge_submitted_measurements(self):
        logger.debug('Purging submitted measurements...')
        deleted_counter = self.purge_measurements(
            filter_args=(Measurement.submitted_at_ts.isnot(None),))
        logger.debug('Purged submitted measurements: %s', deleted_counter)

        return deleted_counter

    def purge_failed_measurements(self):
        logger.debug('Purging failed measurements...')
        created_at_threshold_ts = (
            time.time() - settings.get_worker_client_side_failed_submissions_timeout_seconds())

        deleted_counter = self.purge_measurements(
            filter_args=(Measurement.error_side == ErrorSide.CLIENT.value,
                         Measurement.created_at_ts <= created_at_threshold_ts))
        logger.debug('Purged failed measurements: %s', deleted_counter)

        return deleted_counter

    def get_measurements_number(self):
        with Session(self.engine) as session:
            return session.query(func.count('*')).select_from(Measurement).scalar()


def get_local_storage():
    local_storage = getattr(container, 'local_storage', None)
    if not local_storage:
        container.local_storage = local_storage = LocalStorage()

    return local_storage



import logging

from itertools import chain
from abc import ABC

from packy_agent.worker.loops.base.startable import StartableLoop
from packy_agent.worker.loops.base.stoppable import StoppableLoop
from packy_agent.exceptions import ImproperlyConfiguredError


logger = logging.getLogger(__name__)


def describe_callable(callable_, args, kwargs):
    return '{}({})'.format(
        callable_.__name__,
        ', '.join(chain(map(repr, args), (f'{k}={v !r}' for k, v in kwargs.items()))))


class StartableStoppableLoop(StoppableLoop, StartableLoop, ABC):

    loop_type = 'startable_stoppable'


class CallableLoop(StartableStoppableLoop, ABC):

    loop_type = 'callable'

    def __init__(self, callable_=None, args=(), kwargs=None, is_safe_iteration=True,
                 is_logged_iteration=True, **mkwargs):
        super().__init__(is_safe_iteration=is_safe_iteration,
                         is_logged_iteration=is_logged_iteration, **mkwargs)

        self.callable = callable_
        self.args = args
        self.kwargs = kwargs or {}

    def call(self):
        if self.callable:
            logger.info(f'STARTED actual work: {self.description}')
            result = self.callable(*self.args, **self.kwargs)
            logger.info(f'FINISHED actual work: {self.description}')
        else:
            raise ImproperlyConfiguredError("Provide callable or override 'call' method")

        return result



from schematics import types

from packy_agent.domain_logic.models.schematics.base import BaseMeasurement


class SpeedtestModuleMeasurement(BaseMeasurement):
    target = types.IntType(required=True)
    upload_speed = types.FloatType()
    download_speed = types.FloatType()
    ping = types.FloatType()



import logging
import os
import socket
import select
import time

from ryu.lib.packet.icmp import ICMP_ECHO_REPLY

from contextlib import closing
from packy_agent.domain_logic.modules.common.socket import (
    get_seq_no, TIME_SIZE_BYTES, send_icmp_echo_request, decode_timestamp)
from packy_agent.domain_logic.modules.common.buffer import (
    pop_buffered_icmp_response_receive_time, buffer_icmp_response)
from packy_agent.domain_logic.modules.common.misc import retrieve_ip_and_icmp_packets


logger = logging.getLogger(__name__)


def process_probe_reply(packet_data, seq_no, id_, expected_payload):
    result = retrieve_ip_and_icmp_packets(packet_data)
    if result:
        ip_packet, icmp_packet = result
    else:
        return

    ip_address = ip_packet.src

    logger.debug('Got ICMP packet type:%s code:%s from %s',
                 icmp_packet.type, icmp_packet.code, ip_address)
    if icmp_packet.type != ICMP_ECHO_REPLY:
        logger.debug('Discarded ICMP packet type:{} code:{} from {}'.format(
            icmp_packet.type, icmp_packet.code, ip_address))
        return

    packet_seq = icmp_packet.data.seq
    packet_id = icmp_packet.data.id
    packet_payload = icmp_packet.data.data
    expected_payload_len = len(expected_payload)

    logger.debug(
        'ICMP packet from {} type:{} code:{} has id:{}, seq:{}, payload:{}'.format(
            ip_address, icmp_packet.type, icmp_packet.code, packet_id, packet_seq,
            packet_payload.hex() if packet_payload else packet_payload))

    if packet_seq != seq_no or packet_id != id_:
        return

    if packet_payload:
        packet_payload_len = len(packet_payload)
        if packet_payload_len != expected_payload_len:
            logger.warning('Received payload length ({} bytes) differ from expected (sent) '
                           'length ({} bytes), but we are forgiving'.format(
                packet_payload_len, expected_payload_len))
            min_len = min(packet_payload_len, expected_payload_len)
            if packet_payload[:min_len] != expected_payload[:min_len]:
                logger.debug('Discarded ICMP packet type:{} code:{} from {}: '
                             'payload prefix differs'.format(icmp_packet.type,
                                                             icmp_packet.code, ip_address))
                return
        elif packet_payload != expected_payload:
            logger.debug('Discarded ICMP packet type:{} code:{} from {}: '
                         'payload differs'.format(icmp_packet.type, icmp_packet.code,
                                                  ip_address))
            return

    return packet_payload


def receive_icmp_reply(raw_socket, id_, seq_no, expected_payload, timeout, sent_time):
    expected_payload_len = len(expected_payload)

    time_left = timeout
    while time_left > 0:
        start_time = time.time()
        ready_sockets = select.select((raw_socket,), (), (), time_left)
        receive_time = time.time()  # because by this time the entire packet is already in OS buffer

        if not ready_sockets[0]:  # socket timeout
            break

        time_left -= (receive_time - start_time)

        packet_data, (ip_address, _) = raw_socket.recvfrom(1024 + expected_payload_len)
        packet_payload = process_probe_reply(packet_data, seq_no, id_, expected_payload)
        if packet_payload is None:
            buffer_icmp_response(packet_data, receive_time)
            continue

        if packet_payload and len(packet_payload) >= TIME_SIZE_BYTES:
            sent_time = decode_timestamp(packet_payload[:TIME_SIZE_BYTES])

        buffered_receive_time = pop_buffered_icmp_response_receive_time(packet_data)
        if buffered_receive_time is not None:
            return buffered_receive_time - sent_time
        else:
            return receive_time - sent_time


def ping_once(destination_ip_address, timeout, packet_size_bytes=56):
    icmp_protocol = socket.getprotobyname('icmp')
    try:
        with closing(socket.socket(socket.AF_INET, socket.SOCK_RAW, icmp_protocol)) as raw_socket:
            id_ = os.getpid() & 0xFFFF
            seq_no = get_seq_no('icmp')
            sent_time, payload = send_icmp_echo_request(
                raw_socket, destination_ip_address, id_, seq_no, packet_size_bytes)
            return receive_icmp_reply(raw_socket, id_, seq_no, payload, timeout,
                                      sent_time=sent_time)

    except OSError as ex:
        if ex.errno == 1:  # Operation not permitted
            # TODO(dmu) HIGH: Allow not root pings
            raise OSError(ex.strerror +
                          ' - Note that ICMP messages can only be sent from processes '
                          'running as root.')

        raise



import logging
import time

import certifi
from wampy.peers.clients import Client, WampyError

from packy_agent.utils.wampy import TicketMessageHandler, CustomWebSocket, CustomSecureWebSocket
from packy_agent.configuration.settings import settings
from packy_agent.constants import HEARTBEAT_EVENT

TOPIC_TEMPLATE = 'agent.{event}.{agent_key}'
HEARTBEAT_OPTIONS = {'exclude_me': False}


logger = logging.getLogger(__name__)


class PackyCommunicationClient(Client):

    def __init__(self):
        roles = settings.get_communication_roles()
        url = settings.get_communication_url()

        ticket = settings.get_access_token()
        message_handler = TicketMessageHandler(ticket)

        super().__init__(url=url, roles=roles, message_handler=message_handler,
                         cert_path=certifi.where())

        # TODO(dmu) HIGH: Remove this once https://github.com/noisyboiler/wampy/pull/70/files
        #                 is released
        if self.router.scheme == 'ws':
            self.transport = CustomWebSocket(server_url=self.router.url, ipv=self.router.ipv)
        elif self.router.scheme == 'wss':
            self.transport = CustomSecureWebSocket(
                server_url=self.router.url, ipv=self.router.ipv,
                certificate_path=self.router.certificate)
        else:
            raise WampyError('Network protocol must be "ws" or "wss"')

        self.has_registered_roles = False
        self.handlers = {}

    def set_handler(self, event, handler):
        self.handlers[event] = handler

    def register_roles(self):
        super().register_roles()

        agent_key = settings.get_agent_key()

        for event, handler in self.handlers.items():
            topic = TOPIC_TEMPLATE.format(event=event, agent_key=agent_key)
            self.session._subscribe_to_topic(handler, topic)
            logger.debug("%s subscribed to topic '%s'", self.name, topic)

        self.has_registered_roles = True

    def send_heartbeat(self, agent_key):
        request_ts = time.time()
        self.publish(
            topic=TOPIC_TEMPLATE.format(event=HEARTBEAT_EVENT, agent_key=agent_key),
            message={'request_ts': request_ts},
            options=HEARTBEAT_OPTIONS)

        return request_ts



import os
import sys
import logging

import yaml

from packy_agent.configuration.settings import settings
from packy_agent.configuration.sources.defaults_dict import defaults
from packy_agent.utils.cli import get_base_argument_parser
from packy_agent.utils.logging import configure_logging_basic
from packy_agent.domain_logic.managers.install_and_upgrade import install_and_upgrade_manager
from packy_agent.clients.packy_server import get_packy_server_client


# Use as least as possible from rest of the package because it will run at partially
# installed stage


DEFAULT_LOG_LEVEL = 'INFO'

AGENT_CONFIG_ENVVAR_NAME = 'PACKY_AGENT_CONFIG'
DEFAULT_AGENT_CONFIG_PATH = '/etc/packy-agent.yaml'
CONTROL_SERVER_CONFIG_ENVVAR_NAME = 'PACKY_CONTROL_SERVER_CONFIG'
DEFAULT_CONTROL_SERVER_CONFIG_PATH = '/etc/packy-control-server.yaml'


logger = logging.getLogger(__name__)


def read_control_server_configuration():
    control_server_config_filename = os.getenv(
                CONTROL_SERVER_CONFIG_ENVVAR_NAME, DEFAULT_CONTROL_SERVER_CONFIG_PATH)
    logger.info('Reading Control Server configuration file: %s', control_server_config_filename)
    try:
        with open(control_server_config_filename) as f:
            return yaml.load(f)
    except FileNotFoundError:
        logger.info('Control Server configuration file was not found: %s',
                    control_server_config_filename)
        return


def migrate_console_http_port(control_server_config):
    http_port = ((control_server_config.get('packy') or {}).get('control_server') or {}
                 ).get('http_port')
    if http_port is not None:
        settings_filename = os.getenv('PACKY_AGENT_SETTINGS_FILENAME',
                                      defaults['settings_filename'])
        with open(settings_filename) as f:
            settings_file = yaml.load(f)
            settings_file['console'] = settings_file.get('console') or {}
            settings_file['console']['http_port'] = http_port

        with open(settings_filename, 'w') as f:
            yaml.dump(settings_file, f, default_flow_style=False)


def migrate_flask_secret_key(control_server_config):
    secret_key = (control_server_config.get('flask') or {}).get('SECRET_KEY')
    if secret_key:
        settings.set_console_flask_secret_key(secret_key)


def read_agent_configuration():
    agent_config_filename = os.getenv(AGENT_CONFIG_ENVVAR_NAME, DEFAULT_AGENT_CONFIG_PATH)
    logger.info('Reading Agent configuration file: %s', agent_config_filename)

    try:
        with open(agent_config_filename) as f:
            return yaml.load(f)
    except FileNotFoundError:
        logger.info('Agent configuration file was not found: %s', agent_config_filename)
        return


def migrate_activation_status(agent_configuration):
    inner_config = (agent_configuration.get('packy') or {}).get('agent')
    if not inner_config:
        logger.info('packy.agent key is not found, maybe agent was not activated')
        return

    agent_id = inner_config.get('agent_id')
    if agent_id is None:
        logger.warning('agent_id was not found or empty')
        return

    client_id = inner_config.get('client_id')
    if client_id is None:
        logger.warning('client_id was not found or empty')
        return

    client_secret = inner_config.get('client_secret')
    if client_secret is None:
        logger.warning('client_secret was not found or empty')
        return

    temp_access_token = get_packy_server_client().get_access_token_for_credentials(client_id,
                                                                             client_secret)
    get_packy_server_client().deactivate_agent_legacy(temp_access_token, agent_id)
    install_and_upgrade_manager.activate(access_token=temp_access_token, agent_id=agent_id,
                                         ensure_restarted=False)


def migrate():
    control_server_configuration = read_control_server_configuration()
    if control_server_configuration:
        migrate_console_http_port(control_server_configuration)
        migrate_flask_secret_key(control_server_configuration)

    agent_configuration = read_agent_configuration()
    if agent_configuration:
        migrate_activation_status(agent_configuration)


def entry():
    configure_logging_basic(DEFAULT_LOG_LEVEL)
    parser = get_base_argument_parser(__loader__.name, default_log_level=DEFAULT_LOG_LEVEL)
    args = parser.parse_args()
    configure_logging_basic(args.log_level)

    return migrate()


if __name__ == '__main__':
    sys.exit(entry())



from flask_wtf import FlaskForm
from wtforms import StringField, PasswordField
from wtforms.validators import DataRequired, Email


class LoginForm(FlaskForm):

    email = StringField('E-mail', validators=[DataRequired(), Email()])
    password = PasswordField('Password', validators=[DataRequired()])



from packy_agent.configuration.sources.base import MappingBackedSource
from packy_agent.exceptions import ImproperlyConfiguredError
from packy_agent.configuration.sources.defaults_dict import defaults


def raise_improperly_configured_error(message):
    raise ImproperlyConfiguredError(message)


class ConfigurationDefaults(MappingBackedSource):

    def __init__(self, name=None):
        super().__init__(defaults, name=name or 'Defaults')



import logging

import flask
from flask.views import MethodView

from packy_agent.console.forms.action import ActionForm
from packy_agent.configuration.settings import settings
from packy_agent.domain_logic.managers.reset import reset_manager
from packy_agent.domain_logic.managers.network import network_manager
from packy_agent.utils.auth import activation_and_authentication_required, logout
from packy_agent.console.views.base import smart_redirect


OPERATION_DELAY_SECONDS = 5

logger = logging.getLogger(__name__)


def do_action(action):
    if action == 'full_reset':
        if not settings.is_network_configuration_enabled():
            flask.flash(
                'Network configuration is not allowed. Please, see logs for more information')

        logout()
        reset_manager.full_reset(delay_seconds=OPERATION_DELAY_SECONDS)
        flask.flash(f'Reboot will start in {OPERATION_DELAY_SECONDS} seconds...')
    elif action == 'deactivate':
        logout()
        reset_manager.deactivate(delay_seconds=OPERATION_DELAY_SECONDS)
    elif action == 'network_reset':
        if not settings.is_network_configuration_enabled():
            flask.flash(
                'Network configuration is not allowed. Please, see logs for more information')

        reset_manager.reset_network_configuration(reboot_delay_seconds=OPERATION_DELAY_SECONDS)
        flask.flash(f'Reboot will start in {OPERATION_DELAY_SECONDS} seconds...')
    else:
        raise ValueError('Unknown action: {}'.format(action))


class ResetView(MethodView):

    @activation_and_authentication_required
    def get(self):
        form = ActionForm()
        context = {
            'form': form,
            'active_menu_item': 'reset',
            'is_configuration_restorable': network_manager.is_configuration_restorable()
        }

        return flask.render_template('reset.html', **context)

    @activation_and_authentication_required
    def post(self):
        form = ActionForm()
        if form.validate():
            action = flask.request.form['action']
            do_action(action)
            return smart_redirect('success', 'index', button_text='Continue')

        return flask.render_template('reset.html', form=form)



import time
import logging

from packy_agent.utils.datetime import get_croniter
from packy_agent.worker.loops.base.misc import CallableLoop, describe_callable
from packy_agent.exceptions import ImproperlyConfiguredError


logger = logging.getLogger(__name__)


class ScheduledLoop(CallableLoop):

    loop_type = 'scheduled'

    def __init__(self, schedule=None, callable_=None, args=(), kwargs=None, is_safe_iteration=True,
                 is_logged_iteration=True, **mkwargs):
        super().__init__(callable_=callable_, args=args, kwargs=kwargs,
                         is_safe_iteration=is_safe_iteration,
                         is_logged_iteration=is_logged_iteration, **mkwargs)

        if not schedule:
            raise ImproperlyConfiguredError('Schedule must be provided')

        self.schedule = schedule

        self.croniter = None

    def iteration(self):
        task_description = self.description

        next_moment = self.croniter.get_next()
        current_moment = time.time()
        if next_moment >= current_moment:
            wait_time = next_moment - current_moment
            if wait_time > 0:
                logger.debug(f'WAITING {wait_time:.6g} seconds for next '
                             f'iteration of {task_description} at {next_moment}')
                self.sleep(wait_time)

            self.call()
        else:
            logger.warning(f'MISSED {next_moment:.1f} epoch moment for {task_description}')

    def loop(self):
        self.croniter = get_croniter(self.schedule)
        super().loop()

    @property
    def description(self):
        return describe_callable(self.callable or self.__class__.__name__,
                                 self.args, self.kwargs) + f' ({self.schedule})'
